"""
Port of the AMD Xilinx AIE dialect for programming the AIEs on the AMD Xilinx Versal FPGA architecture.
AIE is a hardened systolic array present in the Versal devices. The dialect describes netlists of AIE
components and it can be lowered to the processor's assembly using the vendor's compiler. A description
of the original dialect can be found here https://xilinx.github.io/mlir-aie/AIEDialect.html
"""

from __future__ import annotations

from collections.abc import Iterable
from enum import auto
from typing import Annotated, Generic

from xdsl.dialects import memref
from xdsl.dialects.builtin import (
    AnyAttr,
    AnyIntegerAttr,
    ArrayAttr,
    Block,
    Float32Type,
    IndexType,
    IntAttr,
    IntegerAttr,
    IntegerType,
    MemRefType,
    ModuleOp,
    NoneAttr,
    Region,
    Signedness,
    StringAttr,
    SymbolRefAttr,
    i32,
)
from xdsl.dialects.func import FuncOp
from xdsl.ir import (
    Attribute,
    AttributeInvT,
    Dialect,
    EnumAttribute,
    Operation,
    OpResult,
    ParametrizedAttribute,
    SSAValue,
    StrEnum,
)
from xdsl.irdl import (
    IRDLOperation,
    ParameterDef,
    attr_def,
    irdl_attr_definition,
    irdl_op_definition,
    operand_def,
    opt_attr_def,
    opt_region_def,
    region_def,
    result_def,
    successor_def,
    var_operand_def,
)
from xdsl.parser import Parser
from xdsl.printer import Printer
from xdsl.traits import (
    HasParent,
    IsTerminator,
    NoTerminator,
    SingleBlockImplicitTerminator,
    SymbolOpInterface,
    SymbolTable,
)
from xdsl.utils.exceptions import VerifyException
from xdsl.utils.hints import isa

i8 = IntegerType(8)

CASCADE_SIZE = 384


class AIEDeviceEnum(StrEnum):
    xcvc1902 = auto()


@irdl_attr_definition
class AIEDeviceAttr(EnumAttribute[AIEDeviceEnum]):
    name = "aie.device_attr"


class ObjectFifoPortEnum(StrEnum):
    Produce = "Produce"
    Consume = "Consume"


@irdl_attr_definition
class ObjectFifoPortAttr(EnumAttribute[ObjectFifoPortEnum]):
    name = "aie.port"


class DMAChannelDirEnum(StrEnum):
    S2MM = "S2MM"
    MM2S = "MM2S"


@irdl_attr_definition
class DMAChannelDirAttr(EnumAttribute[DMAChannelDirEnum]):
    name = "aie.channel_dir"


class WireBundleEnum(StrEnum):
    Core = "Core"
    DMA = "DMA"
    FIFO = "FIFO"
    South = "South"
    West = "West"
    North = "North"
    East = "East"
    PLIO = "PLIO"
    NOC = "NOC"
    Trace = "Trace"


@irdl_attr_definition
class WireBundleAttr(EnumAttribute[WireBundleEnum]):
    name = "aie.wire_bundle"


class LockActionEnum(StrEnum):
    Acquire = "Acquire"
    AcquireGreaterEqual = "AcquireGreaterEqual"
    Release = "Release"


@irdl_attr_definition
class LockActionAttr(EnumAttribute[LockActionEnum]):
    name = "aie.lock_action"


class LockBlockingEnum(StrEnum):
    NonBlocking = "NonBlocking"
    Blocking = "Blocking"


@irdl_attr_definition
class LockBlockingAttr(EnumAttribute[LockActionEnum]):
    name = "aie.lock_blocking"


class BufferTypeEnum(StrEnum):
    A = "A"
    B = "B"


@irdl_attr_definition
class BufferTypeAttr(EnumAttribute[BufferTypeEnum]):
    name = "aie.buffer_type"


@irdl_attr_definition
class ObjectFIFO(Generic[AttributeInvT], ParametrizedAttribute):
    name = "aie.objectfifo"

    buffer: ParameterDef[AttributeInvT]

    @staticmethod
    def from_element_type_and_shape(
        referenced_type: AttributeInvT,
        shape: Iterable[int | IntAttr],
        layout: Attribute = NoneAttr(),
        memory_space: Attribute = NoneAttr(),
    ) -> ObjectFIFO[AttributeInvT]:
        return ObjectFIFO(
            [memref.MemRefType(referenced_type, shape, layout, memory_space)]
        )


@irdl_attr_definition
class ObjectFIFOSubview(Generic[AttributeInvT], ParametrizedAttribute):
    name = "aie.objectfifosubview"

    buffer: ParameterDef[memref.MemRefType[AttributeInvT]]

    @staticmethod
    def from_element_type_and_shape(
        element_type: AttributeInvT,
        shape: Iterable[int | IntAttr],
    ) -> ObjectFIFOSubview[AttributeInvT]:
        return ObjectFIFOSubview([memref.MemRefType(element_type, shape)])


@irdl_op_definition
class SwitchboxOp(IRDLOperation):
    name = "aie.switchbox"

    tile = operand_def(IndexType())
    region = region_def()
    result = result_def(IndexType())

    def __init__(self, tile: Operation | SSAValue, region: Region):
        super().__init__(operands=[tile], regions=[region], result_types=[IndexType()])

    def print(self, printer: Printer):
        printer.print("(")
        printer.print_operand(self.tile)
        printer.print(") ")
        printer.print_region(self.region)

    @classmethod
    def parse(cls, parser: Parser) -> SwitchboxOp:
        parser.parse_characters("(")
        tile = parser.parse_operand()
        parser.parse_characters(")")
        region = parser.parse_region()

        return SwitchboxOp(tile, region)


@irdl_op_definition
class AMSelOp(IRDLOperation):
    name = "aie.amsel"
    arbiterID = attr_def(AnyIntegerAttr)
    msel = attr_def(AnyIntegerAttr)
    result = result_def(IndexType())

    traits = frozenset([HasParent(SwitchboxOp)])

    def __init__(
        self, arbiterID: IntegerAttr[IntegerType], msel: IntegerAttr[IntegerType]
    ):
        super().__init__(
            attributes={"arbiterID": arbiterID, "msel": msel},
            result_types=[IndexType()],
        )

    def verify_(self) -> None:
        if self.arbiterID.type != IntegerType(8):
            raise VerifyException("arbiterID has to be an 8-bit signless integer")
        if self.arbiterID.value.data < 0 or self.arbiterID.value.data > 5:
            raise VerifyException("arbiterID has to be in the range [0-5].")
        if self.msel.type != IntegerType(8):
            raise VerifyException("msel has to be an 8-bit signless integer")
        if self.msel.value.data < 0 or self.arbiterID.value.data > 3:
            raise VerifyException("msel has to be in the range [0-3].")


@irdl_op_definition
class BufferOp(IRDLOperation):
    name = "aie.buffer"
    tile = operand_def(IndexType())
    buffer = result_def(memref.MemRefType)
    sym_name = attr_def(StringAttr)

    def __init__(
        self,
        tile: Operation | SSAValue,
        element_type: Attribute,
        shape: ArrayAttr[AnyIntegerAttr],
        sym_name: StringAttr,
    ):
        buffer_type = memref.MemRefType.from_element_type_and_shape(element_type, shape)
        super().__init__(
            operands=[tile],
            attributes={"sym_name": sym_name},
            result_types=[buffer_type],
        )

    assembly_format = "`(` $tile `)` attr-dict `:` type($buffer)"


@irdl_op_definition
class TileOp(IRDLOperation):
    name = "aie.tile"
    col = attr_def(IntegerAttr[IntegerType])
    row = attr_def(IntegerAttr[IntegerType])
    result = result_def(IndexType())

    def __init__(self, col: IntegerAttr[IntegerType], row: IntegerAttr[IntegerType]):
        super().__init__(
            attributes={"col": col, "row": row}, result_types=[IndexType()]
        )

    def print(self, printer: Printer):
        printer.print("(")
        printer.print(self.col.value.data)
        printer.print(",")
        printer.print(self.row.value.data)
        printer.print(")")

    @classmethod
    def parse(cls, parser: Parser) -> TileOp:
        parser.parse_characters("(")
        col = IntegerAttr.from_int_and_width(parser.parse_integer(), 32)
        parser.parse_characters(",")
        row = IntegerAttr.from_int_and_width(parser.parse_integer(), 32)
        parser.parse_characters(")")

        return TileOp(col, row)


@irdl_op_definition
class ShimMuxOp(IRDLOperation):
    name = "aie.shimmux"

    tile = operand_def(IndexType())

    def __init__(self, tile: Operation | SSAValue):
        super().__init__(operands=[tile], result_types=[IndexType()])


@irdl_op_definition
class ConnectOp(IRDLOperation):
    name = "aie.connect"
    sourceBundle = attr_def(WireBundleAttr)
    sourceChannel = attr_def(AnyIntegerAttr)
    destBundle = attr_def(WireBundleAttr)
    destChannel = attr_def(AnyIntegerAttr)

    traits = frozenset([HasParent(SwitchboxOp, ShimMuxOp)])

    def __init__(
        self,
        sourceBundle: WireBundleAttr,
        sourceChannel: int | AnyIntegerAttr,
        destBundle: WireBundleAttr,
        destChannel: int | AnyIntegerAttr,
    ):
        if isinstance(sourceChannel, int):
            sourceChannel = IntegerAttr(sourceChannel, IntegerType(8))
        if isinstance(destChannel, int):
            destChannel = IntegerAttr(destChannel, IntegerType(8))
        super().__init__(
            attributes={
                "sourceBundle": sourceBundle,
                "sourceChannel": sourceChannel,
                "destBundle": destBundle,
                "destChannel": destChannel,
            }
        )

    def print(self, printer: Printer):
        printer.print(
            "<",
            self.sourceBundle.data,
            ": ",
            self.sourceChannel.value.data,
            ", ",
            self.destBundle.data,
            ": ",
            self.destChannel.value.data,
            ">",
        )

    @classmethod
    def parse(cls, parser: Parser) -> ConnectOp:
        parser.parse_characters("<")
        sourceBundle = WireBundleAttr(WireBundleAttr.parse_parameter(parser))
        parser.parse_characters(":")
        sourceChannel = IntegerAttr.from_int_and_width(parser.parse_integer(), 32)
        parser.parse_characters(",")
        destBundle = WireBundleAttr(WireBundleAttr.parse_parameter(parser))
        parser.parse_characters(":")
        sourceChannel = IntegerAttr.from_int_and_width(parser.parse_integer(), 32)
        parser.parse_characters(">")

        return ConnectOp(sourceBundle, sourceChannel, destBundle, sourceChannel)

    def verify_(self) -> None:
        if self.sourceChannel.type != i32:
            raise VerifyException("sourceChannel has to be a 32-bit signless integer")
        if self.sourceChannel.value.data < 0:
            raise VerifyException("sourceChannel has to be equal or greater than 0")
        if self.destChannel.type != i32:
            raise VerifyException("destChannel has to be a 32-bit signless integer")
        if self.destChannel.value.data < 0:
            raise VerifyException("destChannel has to be equal or greater than 0")


@irdl_op_definition
class CoreOp(IRDLOperation):
    name = "aie.core"
    stackSize = opt_attr_def(IntegerType)
    tile = operand_def(IndexType())
    region = region_def()
    link_with = opt_attr_def(StringAttr)
    result = result_def(IndexType())

    def __init__(
        self,
        stackSize: IntegerAttr[IntegerType] | None,
        tile: Operation | SSAValue,
        region: Region,
        link_with: StringAttr | None = None,
    ):
        super().__init__(
            attributes={"stackSize": stackSize, "link_with": link_with},
            operands=[tile],
            regions=[region],
            result_types=[IndexType()],
        )

    def print(self, printer: Printer):
        printer.print(" (")
        printer.print_operand(self.tile)
        printer.print(") ")
        printer.print_region(self.region)
        if self.link_with is not None:
            printer.print(" { link_with=", self.link_with, " }")

    @classmethod
    def parse(cls, parser: Parser) -> CoreOp:
        parser.parse_characters("(")
        tile = parser.parse_operand()
        parser.parse_characters(")")
        region = parser.parse_region()
        attr_dict = parser.parse_optional_attr_dict()

        link_with = None
        if attr_dict:
            assert isinstance(attr_dict["link_with"], StringAttr)
            link_with = attr_dict["link_with"]

        stackSize = None
        return CoreOp(stackSize, tile, region, link_with)


@irdl_op_definition
class DMABDOp(IRDLOperation):
    name = "aie.dma_bd"
    offset = attr_def(IntegerAttr[IntegerType])
    length = attr_def(IntegerAttr[IntegerType])
    buffer = operand_def(memref.MemRefType)
    dimensions = opt_attr_def(
        IntegerAttr[IntegerType]
    )  # TODO: this should be implemented as a DimTupleArrayAttr: check https://xilinx.github.io/mlir-aie/AIEDialect.html

    def __init__(
        self,
        offset: IntegerAttr[IntegerType],
        length: IntegerAttr[IntegerType],
        dimensions: None | IntegerAttr[IntegerType],
        buffer: Operation | SSAValue,
    ):
        if dimensions is None:
            super().__init__(
                attributes={"offset": offset, "length": length},
                operands=[buffer],
            )
        else:
            super().__init__(
                attributes={
                    "offset": offset,
                    "length": length,
                    "dimensions": dimensions,
                },
                operands=[buffer],
            )

    def print(self, printer: Printer):
        printer.print("(")
        printer.print_operand(self.buffer)
        printer.print(
            ": ",
            self.buffer.type,
            ", ",
            self.offset.value.data,
            ", ",
            self.length.value.data,
            ")",
        )

    @classmethod
    def parse(cls, parser: Parser) -> DMABDOp:
        parser.parse_characters("(")
        buffer = parser.parse_operand()
        parser.parse_characters(":")
        parser.parse_type()
        parser.parse_characters(",")
        offset = IntegerAttr.from_int_and_width(parser.parse_integer(), 32)
        parser.parse_characters(",")
        length = IntegerAttr.from_int_and_width(parser.parse_integer(), 32)
        parser.parse_characters(")")

        return DMABDOp(offset, length, None, buffer)


@irdl_op_definition
class DMABDPACKETOp(IRDLOperation):
    name = "aie.dma_bd_packet"
    packet_type = attr_def(IntegerAttr[IntegerType])
    packet_id = attr_def(IntegerAttr[IntegerType])

    def __init__(
        self, packet_type: IntegerAttr[IntegerType], packet_id: IntegerAttr[IntegerType]
    ):
        super().__init__(
            attributes={"packet_type": packet_type, "packet_id": packet_id}
        )

    def print(self, printer: Printer):
        printer.print("(")
        printer.print(self.packet_type.value.data)
        printer.print(",")
        printer.print(self.packet_id.value.data)
        printer.print(")")

    @classmethod
    def parse(cls, parser: Parser) -> DMABDPACKETOp:
        parser.parse_characters("(")
        packet_type = IntegerAttr.from_int_and_width(parser.parse_integer(), 32)
        parser.parse_characters(",")
        packet_id = IntegerAttr.from_int_and_width(parser.parse_integer(), 32)
        parser.parse_characters(")")

        return DMABDPACKETOp(packet_type, packet_id)


@irdl_op_definition
class MemOp(IRDLOperation):
    name = "aie.mem"

    tile = operand_def(IndexType())
    region = region_def()
    result = result_def(IndexType())

    def __init__(self, tile: Operation | SSAValue, region: Region):
        super().__init__(operands=[tile], result_types=[IndexType()], regions=[region])

    def print(self, printer: Printer):
        printer.print("(")
        printer.print(self.tile)
        printer.print(")")
        printer.print_region(self.region)

    @classmethod
    def parse(cls, parser: Parser) -> MemOp:
        parser.parse_characters("(")
        tile = parser.parse_operand()
        parser.parse_characters(")")
        region = parser.parse_region()

        return MemOp(tile, region)


@irdl_op_definition
class MemTileDMAOp(IRDLOperation):
    name = "aie.memTileDMA"

    tile = operand_def(IndexType())

    def __init__(self, tile: Operation | SSAValue):
        super().__init__(operands=[tile], result_types=[IndexType()])


@irdl_op_definition
class ShimDMAOp(IRDLOperation):
    name = "aie.shimDMA"

    tile = operand_def(IndexType())

    def __init__(self, tile: Operation | SSAValue):
        super().__init__(operands=[tile], result_types=[IndexType()])


@irdl_op_definition
class DMAStartOp(IRDLOperation):
    name = "aie.dma_start"
    channelDir = attr_def(DMAChannelDirAttr)
    channelIndex = attr_def(IntegerAttr[IntegerType])
    dest = successor_def()
    chain = successor_def()

    traits = frozenset(
        [IsTerminator(), HasParent(MemOp, MemTileDMAOp, FuncOp, ShimDMAOp)]
    )

    def __init__(
        self,
        channelDir: Attribute,
        channelIndex: IntegerAttr[IntegerType],
        dest: Block,
        chain: Block,
    ):
        super().__init__(
            attributes={"channelDir": channelDir, "channelIndex": channelIndex},
            successors=[dest, chain],
        )

    def print(self, printer: Printer):
        printer.print(
            "(", self.channelDir.data, ", ", self.channelIndex.value.data, ", "
        )
        printer.print_block_name(self.dest)
        printer.print(", ")
        printer.print_block_name(self.chain)
        printer.print(")")

    @classmethod
    def parse(cls, parser: Parser) -> DMAStartOp:
        parser.parse_characters("(")
        channelDir = DMAChannelDirAttr(DMAChannelDirAttr.parse_parameter(parser))
        parser.parse_characters(",")
        channelIndex = IntegerAttr.from_int_and_width(parser.parse_integer(), 32)
        parser.parse_characters(",")
        dest = parser.parse_successor()
        parser.parse_characters(",")
        chain = parser.parse_successor()
        parser.parse_characters(")")

        return DMAStartOp(channelDir, channelIndex, dest, chain)


@irdl_op_definition
class DebugOp(IRDLOperation):
    name = "aie.debug"
    arg = operand_def(AnyAttr())

    def __init__(self, arg: Operation | SSAValue):
        super().__init__(operands=[arg])


@irdl_op_definition
class DeviceOp(IRDLOperation):
    name = "aie.device"

    region = opt_region_def()

    device = attr_def(AIEDeviceAttr)
    traits = frozenset([SymbolTable(), NoTerminator(), HasParent(ModuleOp)])

    def __init__(self, device: AIEDeviceAttr, region: Region):
        super().__init__(attributes={"device": device}, regions=[region])

    def print(self, printer: Printer):
        printer.print("(")
        device_str = "xcvc1902" if self.device.data == AIEDeviceEnum.xcvc1902 else ""
        printer.print(device_str)
        printer.print(") ")
        if self.region is not None:
            printer.print_region(self.region)

    @classmethod
    def parse(cls, parser: Parser) -> DeviceOp:
        parser.parse_characters("(")

        device = AIEDeviceAttr(AIEDeviceAttr.parse_parameter(parser))
        parser.parse_characters(")")
        region = parser.parse_region()

        return DeviceOp(device, region)


@irdl_op_definition
class ExternalBufferOp(IRDLOperation):
    name = "aie.external_buffer"

    sym_name = attr_def(StringAttr)
    buffer = result_def(memref.MemRefType)

    def __init__(
        self,
        sym_name: str,
        shape: ArrayAttr[AnyIntegerAttr],
        element_type: Attribute,
    ):
        super().__init__(
            attributes={"sym_name": StringAttr(sym_name)},
            result_types=[
                memref.MemRefType.from_element_type_and_shape(element_type, shape)
            ],
        )

    assembly_format = "attr-dict `:` type($buffer)"


@irdl_op_definition
class FlowOp(IRDLOperation):
    name = "aie.flow"

    sourceBundle = attr_def(WireBundleAttr)
    sourceChannel = attr_def(IntegerAttr[IntegerType])
    destBundle = attr_def(WireBundleAttr)
    destChannel = attr_def(IntegerAttr[IntegerType])
    source = operand_def(IndexType())
    dest = operand_def(IndexType())

    def __init__(
        self,
        sourceBundle: WireBundleAttr,
        sourceChannel: IntegerAttr[IntegerType],
        destBundle: WireBundleAttr,
        destChannel: IntegerAttr[IntegerType],
        source: Operation | SSAValue,
        dest: Operation | SSAValue,
    ):
        super().__init__(
            attributes={
                "sourceBundle": sourceBundle,
                "sourceChannel": sourceChannel,
                "destBundle": destBundle,
                "destChannel": destChannel,
            },
            operands=[source, dest],
        )

    def print(self, printer: Printer):
        printer.print("(")
        printer.print(self.source)
        printer.print(",")
        printer.print(self.sourceBundle.data)
        printer.print(":")
        printer.print(self.sourceChannel.value.data)
        printer.print(",")
        printer.print(self.dest)
        printer.print(",")
        printer.print(self.destBundle.data)
        printer.print(":")
        printer.print(self.destChannel.value.data)
        printer.print(")")

    @classmethod
    def parse(cls, parser: Parser) -> FlowOp:
        parser.parse_characters("(")
        source = parser.parse_operand()
        parser.parse_characters(",")
        sourceBundle = WireBundleAttr(WireBundleAttr.parse_parameter(parser))
        parser.parse_characters(":")
        sourceChannel = IntegerAttr.from_int_and_width(parser.parse_integer(), 32)
        parser.parse_characters(",")
        dest = parser.parse_operand()
        parser.parse_characters(",")
        destBundle = WireBundleAttr(WireBundleAttr.parse_parameter(parser))
        parser.parse_characters(":")
        destChannel = IntegerAttr.from_int_and_width(parser.parse_integer(), 32)
        parser.parse_characters(")")

        return FlowOp(
            sourceBundle, sourceChannel, destBundle, destChannel, source, dest
        )


@irdl_op_definition
class GetCascadeOp(IRDLOperation):
    name = "getCascade"

    traits = frozenset([HasParent(CoreOp)])

    def __init__(self):
        super().__init__(result_types=[IntegerType(CASCADE_SIZE)])


@irdl_op_definition
class GetStreamOp(IRDLOperation):
    name = "aie.getStream"

    channel = operand_def(i32)
    result = result_def(IntegerType)

    traits = frozenset([HasParent(CoreOp)])

    def __init__(self, channel: Operation | SSAValue):
        super().__init__(
            operands=[channel],
            result_types=[i32],  # TODO: the result can be of several types.
        )


@irdl_op_definition
class LockOp(IRDLOperation):
    name = "aie.lock"

    tile = operand_def(IndexType())
    lockID = opt_attr_def(IntegerAttr[IntegerType])
    init = opt_attr_def(IntegerAttr[IntegerType])
    result = result_def(IndexType())
    sym_name = opt_attr_def(StringAttr)

    def __init__(
        self,
        lockID: IntegerAttr[IntegerType],
        init: IntegerAttr[IntegerType] | None,
        tile: Operation | SSAValue,
        sym_name: StringAttr | None,
    ):
        super().__init__(
            attributes={"lockID": lockID, "init": init, "sym_name": sym_name},
            operands=[tile],
            result_types=[IndexType()],
        )

    def print(self, printer: Printer):
        printer.print("(")
        printer.print(self.tile)
        if self.lockID:
            printer.print(",")
            printer.print(self.lockID.value.data)
            printer.print(")")
        if self.sym_name:
            printer.print("{")
            printer.print("sym_name = ")
            printer.print(self.sym_name)
            printer.print("}")

    @classmethod
    def parse(cls, parser: Parser) -> LockOp:
        parser.parse_characters("(")
        tile = parser.parse_operand()
        parser.parse_characters(",")
        lockID = IntegerAttr.from_int_and_width(parser.parse_integer(), 32)
        init = parser.parse_optional_integer()
        parser.parse_characters(")")
        sym_name = None
        attr_dict = parser.parse_optional_attr_dict()
        if attr_dict:
            assert isinstance(attr_dict["sym_name"], StringAttr)
            sym_name = attr_dict["sym_name"]

        if init:
            init = IntegerAttr.from_int_and_width(init, 32)

        return LockOp(
            lockID, None, tile, sym_name
        )  # TODO: argument should be init but Pyright fails


@irdl_op_definition
class MasterSetOp(IRDLOperation):
    name = "aie.masterset"

    destBundle = attr_def(WireBundleAttr)
    destChannel = attr_def(IntegerAttr[IntegerType])
    amsels = operand_def(IndexType())

    traits = frozenset([HasParent(SwitchboxOp)])

    def __init__(
        self,
        destBundle: WireBundleAttr,
        destChannel: IntegerAttr[IntegerType],
        amsels: Operation | SSAValue,
    ):
        super().__init__(
            attributes={"destBundle": destBundle, "destChannel": destChannel},
            operands=[amsels],
            result_types=[IndexType()],
        )


@irdl_op_definition
class NextBDOp(IRDLOperation):
    name = "aie.next_bd"

    dest = successor_def()

    traits = frozenset(
        [HasParent(MemOp, MemTileDMAOp, FuncOp, ShimDMAOp), IsTerminator()]
    )

    def __init__(self, dest: Block):
        super().__init__(successors=[dest])

    def print(self, printer: Printer):
        printer.print(" ")
        printer.print_block_name(self.dest)

    @classmethod
    def parse(cls, parser: Parser) -> NextBDOp:
        dest = parser.parse_successor()

        return NextBDOp(dest)


@irdl_op_definition
class ObjectFifoAcquireOp(IRDLOperation):
    name = "aie.objectfifo.acquire"

    port = attr_def(ObjectFifoPortAttr)
    size = attr_def(IntegerAttr[IntegerType])
    object_fifo = attr_def(SymbolRefAttr)

    result: OpResult = result_def(ObjectFIFOSubview)

    def __init__(
        self,
        port: ObjectFifoPortAttr,
        size: IntegerAttr[IntegerType],
        object_fifo: str | SymbolRefAttr,
        shape: Iterable[int | IntAttr],
        element_type: Attribute,
    ):
        if isinstance(object_fifo, str):
            object_fifo = SymbolRefAttr(object_fifo)

        result_subview = ObjectFIFOSubview.from_element_type_and_shape(
            element_type, shape
        )
        super().__init__(
            attributes={"port": port, "size": size, "object_fifo": object_fifo},
            result_types=[result_subview],
        )

    def print(self, printer: Printer):
        printer.print(f" @{self.object_fifo.root_reference.data}")
        printer.print(f"( {self.port.data}, {self.size.value.data} )")
        printer.print(" : !aie.objectfifosubview<")
        assert isa(self.result.type, ObjectFIFOSubview[Attribute])
        printer.print(self.result.type.buffer)
        printer.print(">")

    @classmethod
    def parse(cls, parser: Parser) -> ObjectFifoAcquireOp:
        object_fifo = SymbolRefAttr(parser.parse_symbol_name())
        parser.parse_characters("(")
        port = ObjectFifoPortAttr(ObjectFifoPortAttr.parse_parameter(parser))

        parser.parse_characters(",")
        size = IntegerAttr.from_int_and_width(parser.parse_integer(), 32)
        parser.parse_characters(")")
        parser.parse_characters(":")
        parser.parse_characters("!aie.objectfifosubview")
        parser.parse_characters("<")
        ofifo_type = parser.parse_type()
        parser.parse_characters(">")
        assert isa(ofifo_type, MemRefType[Attribute])
        shape = ofifo_type.shape
        element_type = ofifo_type.element_type

        return ObjectFifoAcquireOp(port, size, object_fifo, shape, element_type)


@irdl_op_definition
class ObjectFifoRegisterExternalBuffersOp(IRDLOperation):
    name = "aie.objectfifo.register_external_buffers"

    tile = operand_def(IndexType())
    externalBuffers = operand_def(memref.MemRefType)
    object_fifo = attr_def(SymbolRefAttr)

    traits = frozenset([HasParent(DeviceOp)])

    def __init__(
        self,
        tile: Operation | SSAValue,
        externalBuffers: Operation | SSAValue,
        object_fifo: str | SymbolRefAttr,
    ):
        if isinstance(object_fifo, str):
            object_fifo = SymbolRefAttr(object_fifo)

        super().__init__(
            operands=[tile, externalBuffers], attributes={"object_fifo": object_fifo}
        )

    def print(self, printer: Printer):
        assert isinstance(self.object_fifo, SymbolRefAttr)
        printer.print(
            f" {self.object_fifo} (",
            self.tile,
            ", {",
            self.externalBuffers,
            "}) : (",
            self.externalBuffers.type,
            ")",
        )

    @classmethod
    def parse(cls, parser: Parser) -> ObjectFifoRegisterExternalBuffersOp:
        object_fifo = SymbolRefAttr(parser.parse_symbol_name())
        parser.parse_characters("(")
        tile = parser.parse_operand()
        parser.parse_characters(",")
        parser.parse_characters("{")
        external_buffers = parser.parse_operand()
        parser.parse_characters("}")
        parser.parse_characters(")")
        parser.parse_characters(":")
        parser.parse_characters("(")
        parser.parse_type()
        parser.parse_characters(")")

        return ObjectFifoRegisterExternalBuffersOp(tile, external_buffers, object_fifo)


@irdl_op_definition
class ObjectFIFOSubviewAccessOp(IRDLOperation):
    name = "aie.objectfifo.subview.access"

    index = attr_def(IntegerAttr[IntegerType])
    subview = operand_def(ObjectFIFOSubview[Attribute])
    output = result_def(memref.MemRefType)

    def __init__(self, index: IntegerAttr[IntegerType], subview: Operation | SSAValue):
        assert isinstance(subview, ObjectFifoAcquireOp)
        assert isa(subview.result.type, ObjectFIFOSubview[Attribute])
        subview.result.type.buffer
        result_type = memref.MemRefType(
            subview.result.type.buffer.element_type, subview.result.type.buffer.shape
        )
        super().__init__(
            attributes={"index": index}, operands=[subview], result_types=[result_type]
        )

    def print(self, printer: Printer):
        assert isa(self.subview.type, ObjectFIFOSubview[Attribute])
        printer.print(" ")
        printer.print_operand(self.subview)
        printer.print("[", self.index.value.data, "] : ")
        printer.print(
            "!aie.objectfifosubview<",
            self.subview.type.buffer,
            "> -> ",
            self.subview.type.buffer,
        )

    @classmethod
    def parse(cls, parser: Parser) -> ObjectFIFOSubviewAccessOp:
        subview = parser.parse_operand()
        if isinstance(subview, OpResult):
            subview = subview.op
        parser.parse_characters("[")
        index = IntegerAttr.from_int_and_width(parser.parse_integer(), 32)
        parser.parse_characters("]")
        parser.parse_characters(":")
        parser.parse_characters("!aie.objectfifosubview")
        parser.parse_characters("<")
        parser.parse_type()  # subview type
        parser.parse_characters(">")
        parser.parse_characters("->")
        parser.parse_type()  # return type

        return ObjectFIFOSubviewAccessOp(index, subview)


@irdl_op_definition
class createObjectFifo(IRDLOperation):
    name = "aie.objectfifo"

    elemNumber = attr_def(IntegerAttr[IntegerType])
    producerTile = operand_def(IndexType())
    consumerTiles = var_operand_def(IndexType())
    sym_name = attr_def(StringAttr)
    object_fifo = attr_def(ObjectFIFO[Attribute])

    traits = frozenset([SymbolOpInterface(), HasParent(DeviceOp)])

    def __init__(
        self,
        elemNumber: IntegerAttr[IntegerType],
        producerTile: Operation | SSAValue,
        consumerTiles: list[Operation | SSAValue],
        referenced_type: Attribute,
        shape: Iterable[int | IntAttr],
        name: str,
    ):
        object_fifo = ObjectFIFO.from_element_type_and_shape(referenced_type, shape)
        super().__init__(
            attributes={
                "elemNumber": elemNumber,
                "object_fifo": object_fifo,
                "sym_name": StringAttr(name),
            },
            operands=[producerTile, consumerTiles],
        )

    def print(self, printer: Printer):
        printer.print(" @", self.sym_name.data, "( ", self.producerTile, ", { ")
        for i in range(len(self.consumerTiles) - 1):
            printer.print(self.consumerTiles[i], ", ")

        printer.print(self.consumerTiles[-1], "}, ")
        printer.print(self.elemNumber)

        printer.print(
            ") : !aie.objectfifo<",
            self.object_fifo.buffer,
            ">",
        )

    @classmethod
    def parse(cls, parser: Parser) -> createObjectFifo:
        name = parser.parse_symbol_name().data
        parser.parse_characters("(")
        producerTile = parser.parse_operand()
        parser.parse_characters(",")
        parser.parse_characters("{")
        consumerTiles: list[Operation | SSAValue] = []
        consumerTiles.append(parser.parse_operand())
        while not parser.parse_optional_characters("}"):
            parser.parse_characters(",")
            consumerTiles.append(parser.parse_operand())

        parser.parse_characters(",")
        elemNumber = IntegerAttr.from_int_and_width(parser.parse_integer(), 32)
        parser.parse_characters(":")
        parser.parse_type()
        parser.parse_characters(")")
        parser.parse_characters(":")
        parser.parse_characters("!aie.objectfifo")
        parser.parse_characters("<")
        objectfifo_type = parser.parse_type()
        parser.parse_characters(">")

        assert isa(objectfifo_type, MemRefType[Attribute])
        shape = objectfifo_type.shape
        referenced_type = objectfifo_type.element_type

        object_fifo = createObjectFifo(
            elemNumber, producerTile, consumerTiles, referenced_type, shape, name
        )

        return object_fifo


@irdl_op_definition
class ObjectFIFOReleaseOp(IRDLOperation):
    name = "aie.objectfifo.release"

    port = attr_def(ObjectFifoPortAttr)
    size = attr_def(IntegerAttr[IntegerType])
    object_fifo = attr_def(SymbolRefAttr)

    def __init__(
        self,
        port: ObjectFifoPortAttr,
        size: IntegerAttr[IntegerType],
        object_fifo: str | SymbolRefAttr,
    ):
        if isinstance(object_fifo, str):
            object_fifo = SymbolRefAttr(object_fifo)

        super().__init__(
            attributes={"port": port, "size": size, "object_fifo": object_fifo}
        )

    def print(self, printer: Printer):
        printer.print(
            " ",
            self.object_fifo,
            " (",
            "Produce" if self.port.data == ObjectFifoPortEnum.Produce else "Consume",
            ", ",
            self.size.value.data,
            ")",
        )

    @classmethod
    def parse(cls, parser: Parser) -> ObjectFIFOReleaseOp:
        object_fifo = SymbolRefAttr(parser.parse_symbol_name())
        parser.parse_characters("(")
        port = ObjectFifoPortAttr(ObjectFifoPortAttr.parse_parameter(parser))
        parser.parse_characters(",")

        size = IntegerAttr.from_int_and_width(parser.parse_integer(), 32)

        parser.parse_characters(")")

        return ObjectFIFOReleaseOp(port, size, object_fifo)


@irdl_op_definition
class PLIOOp(IRDLOperation):
    name = "aie.plio"

    col = attr_def(IntegerAttr[IntegerType])

    def __init__(self, col: IntegerAttr[IntegerType]):
        super().__init__(attributes={"col": col}, result_types=[IndexType()])


@irdl_op_definition
class EndOp(IRDLOperation):
    name = "aie.end"

    def __init__(self):
        super().__init__()

    traits = frozenset([IsTerminator()])

    assembly_format = "attr-dict"


@irdl_op_definition
class PacketFlowOp(IRDLOperation):
    name = "aie.packet_flow"

    ID = attr_def(IntegerAttr[IntegerType])
    region = region_def()

    traits = frozenset([SingleBlockImplicitTerminator(EndOp)])

    def __init__(self, ID: IntegerAttr[IntegerType], region: Region):
        super().__init__(attributes={"ID": ID}, regions=[region])

    def print(self, printer: Printer):
        printer.print("(", f"0x{self.ID.value.data:X}", ") ")
        printer.print_region(self.region)

    @classmethod
    def parse(cls, parser: Parser) -> PacketFlowOp:
        parser.parse_characters("(")
        id = IntegerAttr.from_int_and_width(parser.parse_integer(), 32)
        parser.parse_characters(")")
        region = parser.parse_region()

        return PacketFlowOp(id, region)


@irdl_op_definition
class PacketDestOp(IRDLOperation):
    name = "aie.packet_dest"

    bundle = attr_def(WireBundleAttr)
    channel = attr_def(IntegerAttr[IntegerType])

    tile = operand_def(IndexType())

    traits = frozenset([HasParent(PacketFlowOp)])

    def __init__(
        self,
        bundle: WireBundleAttr,
        channel: IntegerAttr[IntegerType],
        tile: Operation | SSAValue,
    ):
        super().__init__(
            attributes={"bundle": bundle, "channel": channel}, operands=[tile]
        )

    def print(self, printer: Printer):
        printer.print("<")
        printer.print(self.tile)
        printer.print(", ")
        printer.print(self.bundle.data)
        printer.print(" : ")
        printer.print(self.channel.value.data)
        printer.print(">")

    @classmethod
    def parse(cls, parser: Parser) -> PacketDestOp:
        parser.parse_characters("<")
        tile = parser.parse_operand()
        parser.parse_characters(",")
        bundle_enum = WireBundleAttr.parse_parameter(parser)
        bundle = WireBundleAttr(bundle_enum)
        parser.parse_characters(":")
        channel = IntegerAttr.from_int_and_width(parser.parse_integer(), 32)
        parser.parse_characters(">")

        return PacketDestOp(bundle, channel, tile)


@irdl_op_definition
class PacketRulesOp(IRDLOperation):
    name = "aie.packetrules"

    sourceBundle = attr_def(WireBundleAttr)
    sourceChannel = attr_def(IntegerAttr[IntegerType])

    def __init__(
        self, sourceBundle: WireBundleAttr, sourceChannel: IntegerAttr[IntegerType]
    ):
        super().__init__(
            attributes={"sourceBundle": sourceBundle, "sourceChannel": sourceChannel}
        )


@irdl_op_definition
class PacketRuleOp(IRDLOperation):
    name = "aie.rule"

    mask = attr_def(IntegerAttr[IntegerType])
    value = attr_def(IntegerAttr[IntegerType])

    amsel = operand_def(IndexType())

    traits = frozenset([HasParent(PacketRulesOp)])

    def __init__(
        self,
        mask: IntegerAttr[IntegerType],
        value: IntegerAttr[IntegerType],
        amsel: Operation | SSAValue,
    ):
        super().__init__(attributes={"mask": mask, "value": value}, operands=[amsel])


@irdl_op_definition
class PacketSourceOp(IRDLOperation):
    name = "aie.packet_source"

    bundle = attr_def(WireBundleAttr)
    channel = attr_def(IntegerAttr[IntegerType])
    tile = operand_def(IndexType())

    traits = frozenset([HasParent(PacketFlowOp)])

    def __init__(
        self,
        bundle: WireBundleAttr,
        channel: IntegerAttr[IntegerType],
        tile: Operation | SSAValue,
    ):
        super().__init__(
            attributes={"bundle": bundle, "channel": channel}, operands=[tile]
        )

    def print(self, printer: Printer):
        printer.print("<")
        printer.print(self.tile)
        printer.print(", ")
        printer.print(self.bundle.data)
        printer.print(" : ")
        printer.print(self.channel.value.data)
        printer.print(">")

    @classmethod
    def parse(cls, parser: Parser) -> PacketSourceOp:
        parser.parse_characters("<")
        tile = parser.parse_operand()
        parser.parse_characters(",")
        bundle_enum = WireBundleAttr.parse_parameter(parser)
        bundle = WireBundleAttr(bundle_enum)
        parser.parse_characters(":")
        channel = IntegerAttr.from_int_and_width(parser.parse_integer(), 32)
        parser.parse_characters(">")

        return PacketSourceOp(bundle, channel, tile)


@irdl_op_definition
class PutCascade(IRDLOperation):
    name = "aie.putCascade"

    cascadeValue = operand_def(IntegerType(CASCADE_SIZE))

    traits = frozenset([HasParent(CoreOp)])

    def __init__(self, cascadeValue: Operation | SSAValue):
        super().__init__(operands=[cascadeValue])


@irdl_op_definition
class PutStream(IRDLOperation):
    name = "aie.putStream"

    channel = operand_def(i32)
    streamValue = operand_def(
        IntegerType(32, Signedness.SIGNLESS)
        or Float32Type
        or IntegerType(128, Signedness.SIGNLESS)
    )

    traits = frozenset([HasParent(CoreOp)])

    def __init__(
        self, channel: Operation | SSAValue, streamValue: Operation | SSAValue
    ):
        super().__init__(operands=[channel, streamValue])


@irdl_op_definition
class ShimDMAAllocationOp(IRDLOperation):
    name = "aie.shimDMAAllocation"

    sym_name = attr_def(StringAttr)
    channelDir = attr_def(IntegerAttr[Annotated[IntegerType, i32]])
    channelIndex = attr_def(IntegerAttr[IntegerType])
    col = attr_def(IntegerAttr[IntegerType])

    traits = frozenset([HasParent(DeviceOp)])

    def __init__(
        self,
        sym_name: StringAttr,
        channelDir: IntegerAttr[IntegerType],
        channelIndex: IntegerAttr[IntegerType],
        col: IntegerAttr[IntegerType],
    ):
        super().__init__(
            attributes={
                "sym_name": sym_name,
                "channelDir": channelDir,
                "channelIndex": channelIndex,
                "col": col,
            }
        )


@irdl_op_definition
class ShimSwitchBoxOp(IRDLOperation):
    name = "aie.shimswitchbox"

    col = attr_def(IntegerAttr[IntegerType])

    def __init__(self, col: IntegerAttr[IntegerType]):
        super().__init__(attributes={"col": col}, result_types=[IndexType()])


@irdl_op_definition
class UseLockOp(IRDLOperation):
    name = "aie.use_lock"

    value = opt_attr_def(IntegerAttr[IntegerType])
    action = attr_def(LockActionAttr)
    blocking = opt_attr_def(LockBlockingAttr)
    lock = operand_def(IndexType())

    def __init__(
        self,
        value: IntegerAttr[IntegerType] | None,
        action: LockActionAttr,
        blocking: LockBlockingAttr | None,
        lock: Operation | SSAValue,
    ):
        super().__init__(
            attributes={"value": value, "action": action, "blocking": blocking},
            operands=[lock],
        )

    def print(self, printer: Printer):
        printer.print("(")
        printer.print_operand(self.lock)
        printer.print(", ")
        printer.print(self.action.data)
        printer.print(", ")
        if self.value:
            printer.print(self.value.value.data)
        printer.print(")")

    @classmethod
    def parse(cls, parser: Parser) -> UseLockOp:
        parser.parse_characters("(")
        lock = parser.parse_operand()
        parser.parse_characters(",")
        action = LockActionAttr(LockActionAttr.parse_parameter(parser))

        value = None
        if parser.parse_optional_characters(","):
            value = IntegerAttr.from_int_and_width(parser.parse_integer(), 32)

        blocking = None
        if parser.parse_optional_characters(","):
            blocking = LockBlockingAttr(LockBlockingAttr.parse_parameter(parser))
        parser.parse_characters(")")

        return UseLockOp(value, action, blocking, lock)


@irdl_op_definition
class WireOp(IRDLOperation):
    name = "aie.wire"

    sourceBundle = attr_def(WireBundleAttr)
    destBundle = attr_def(WireBundleAttr)
    source = operand_def(IndexType())
    dest = operand_def(IndexType())

    def __init__(
        self,
        sourceBundle: WireBundleAttr,
        destBundle: WireBundleAttr,
        source: Operation | SSAValue,
        dest: Operation | SSAValue,
    ):
        super().__init__(
            attributes={"sourceBundle": sourceBundle, "destBundle": destBundle},
            operands=[source, dest],
        )


AIE = Dialect(
    "aie",
    [
        AMSelOp,
        BufferOp,
        TileOp,
        ConnectOp,
        CoreOp,
        DMABDOp,
        DMABDPACKETOp,
        DMAStartOp,
        DebugOp,
        DeviceOp,
        ExternalBufferOp,
        FlowOp,
        GetCascadeOp,
        GetStreamOp,
        LockOp,
        MasterSetOp,
        MemOp,
        MemTileDMAOp,
        NextBDOp,
        ObjectFifoAcquireOp,
        ObjectFifoRegisterExternalBuffersOp,
        ObjectFIFOSubviewAccessOp,
        createObjectFifo,
        ObjectFIFOReleaseOp,
        PLIOOp,
        PacketDestOp,
        PacketFlowOp,
        PacketRuleOp,
        PacketRulesOp,
        PacketSourceOp,
        PutCascade,
        PutStream,
        ShimDMAOp,
        ShimMuxOp,
        ShimSwitchBoxOp,
        SwitchboxOp,
        UseLockOp,
        WireOp,
        EndOp,
    ],
    [
        WireBundleAttr,
        ObjectFIFO,
        ObjectFIFOSubview,
    ],
)



from typing import Any, TypeVar

from typing_extensions import Protocol

from xdsl.dialects import stream
from xdsl.interpreter import (
    Interpreter,
    InterpreterFunctions,
    PythonValues,
    impl,
    register_impls,
)

T = TypeVar("T")
TCov = TypeVar("TCov", covariant=True)
TCon = TypeVar("TCon", contravariant=True)


class ReadableStream(Protocol[TCov]):
    def read(self) -> TCov:
        raise NotImplementedError()


class WritableStream(Protocol[TCon]):
    def write(self, value: TCon) -> None:
        raise NotImplementedError()


@register_impls
class StreamFunctions(InterpreterFunctions):
    @impl(stream.ReadOp)
    def run_read(
        self, interpreter: Interpreter, op: stream.ReadOp, args: tuple[Any, ...]
    ) -> PythonValues:
        (stream,) = args
        stream: ReadableStream[Any] = stream
        return (stream.read(),)

    @impl(stream.WriteOp)
    def run_write(
        self, interpreter: Interpreter, op: stream.WriteOp, args: tuple[Any, ...]
    ) -> PythonValues:
        (value, stream) = args
        stream: WritableStream[Any] = stream
        stream.write(value)
        return ()



import pytest

from xdsl.utils.exceptions import PassPipelineParseError
from xdsl.utils.parse_pipeline import (
    PipelineLexer,
    Token,
    parse_pipeline,
)

Kind = Token.Kind

generator = PipelineLexer._generator  # pyright: ignore[reportPrivateUsage]


def test_pass_lexer():
    tokens = list(
        generator(
            'pass-1,pass-2{arg1=1 arg2=test arg3="test-str" arg-4=-34.4e-12 no-val-arg},pass-3'
        )
    )

    assert [t.kind for t in tokens] == [
        Kind.IDENT, Kind.COMMA,  # pass-1,
        Kind.IDENT, Kind.L_BRACE,  # pass-2{
        Kind.IDENT, Kind.EQUALS, Kind.NUMBER, Kind.SPACE,  # arg1=1
        Kind.IDENT, Kind.EQUALS, Kind.IDENT, Kind.SPACE,  # arg2=test
        Kind.IDENT, Kind.EQUALS, Kind.STRING_LIT, Kind.SPACE,  # arg3="test-str"
        Kind.IDENT, Kind.EQUALS, Kind.NUMBER, Kind.SPACE,  # arg-4=-34.4e-12
        Kind.IDENT,  # no-val-arg
        Kind.R_BRACE, Kind.COMMA,  # },
        Kind.IDENT,  # pass-3
        Kind.EOF,
    ]  # fmt: skip

    assert tokens[-2].span.text == "pass-3"
    assert tokens[1].span.text == ","
    assert tokens[3].span.text == "{"
    assert tokens[18].span.text == "-34.4e-12"

    assert len(list(generator(""))) == 1


def test_pass_lex_errors():
    with pytest.raises(PassPipelineParseError, match="Unknown token"):
        list(generator("pass-1["))

    with pytest.raises(PassPipelineParseError, match="Unknown token"):
        list(generator("pass-1{thing$=1}"))


@pytest.mark.parametrize(
    "input_str, pass_name, pass_arg_names",
    (
        ("pass-1,", "pass-1", set[str]()),
        ("pass-1{}", "pass-1", set[str]()),
        ("pass-1{arg1=true arg2}", "pass-1", {"arg1", "arg2"}),
        ("pass-1{arg2 arg1=false}", "pass-1", {"arg1", "arg2"}),
    ),
)
def test_pass_parser_argument_dict_edge_cases(
    input_str: str, pass_name: str, pass_arg_names: set[str]
):
    """
    This test checks edge-cases in the parsing code.
    """
    passes = list(parse_pipeline(input_str))
    assert len(passes) == 1
    assert passes[0].name == pass_name
    assert set(passes[0].args.keys()) == pass_arg_names


@pytest.mark.parametrize(
    "input_str",
    (
        ("pass-1{"),
        ("pass-1{arg1,arg2}"),
        ("pass-1{arg1=arg2=arg3}"),
        ("pass-1{ }"),
    ),
)
def test_pass_parser_cases_fail(input_str: str):
    with pytest.raises(PassPipelineParseError):
        list(parse_pipeline(input_str))


def test_pass_parse_errors():
    """
    This test triggers all parse errors in the parser in the same order they appear
    in the source file.
    """
    with pytest.raises(PassPipelineParseError, match="Expected pass name here"):
        # numbers are not valid pass names!
        list(parse_pipeline("1"))

    with pytest.raises(
        PassPipelineParseError, match="Expected a comma or pass arguments here"
    ):
        list(parse_pipeline("pass-1="))

    with pytest.raises(
        PassPipelineParseError, match="Expected a comma after pass argument dict here"
    ):
        list(parse_pipeline("pass-1{arg1=1}="))

    with pytest.raises(PassPipelineParseError, match="Expected argument name here"):
        # numbers are not valid argument names
        list(parse_pipeline("pass-1{1=1}"))

    with pytest.raises(
        PassPipelineParseError,
        match="Expected equals, space or end of arguments here",
    ):
        # we don't support the case where there is no `=` after the arg name (yet)
        list(parse_pipeline("pass-1{arg1{1}"))

    with pytest.raises(
        PassPipelineParseError,
        match="Malformed pass arguments, expected either a space or `}` here",
    ):
        list(parse_pipeline("pass-1{arg1=1=}"))

    with pytest.raises(PassPipelineParseError, match="Unknown argument value"):
        list(parse_pipeline("pass-1{arg1={}}"))



from __future__ import annotations

from collections.abc import Sequence
from typing import cast

from typing_extensions import Self

from xdsl.dialects import riscv, stream
from xdsl.dialects.builtin import (
    IntAttr,
    IntegerAttr,
    IntegerType,
    Signedness,
    UnrealizedConversionCastOp,
)
from xdsl.dialects.riscv import (
    AssemblyInstructionArg,
    IntRegisterType,
    RdRsImmIntegerOperation,
    RdRsRsOperation,
    Registers,
    RISCVInstruction,
    RISCVOp,
    UImm5Attr,
)
from xdsl.dialects.utils import (
    AbstractYieldOperation,
    parse_assignment,
    print_assignment,
)
from xdsl.ir import Attribute, Block, Dialect, Operation, Region, SSAValue
from xdsl.irdl import (
    IRDLOperation,
    attr_def,
    irdl_op_definition,
    operand_def,
    prop_def,
    region_def,
    result_def,
    traits_def,
    var_operand_def,
    var_result_def,
)
from xdsl.parser import Parser, UnresolvedOperand
from xdsl.pattern_rewriter import RewritePattern
from xdsl.printer import Printer
from xdsl.traits import (
    HasCanonicalisationPatternsTrait,
    HasParent,
    IsTerminator,
    Pure,
    SingleBlockImplicitTerminator,
    ensure_terminator,
)
from xdsl.utils.exceptions import VerifyException

# region Snitch Extensions


class ScfgwOpHasCanonicalizationPatternsTrait(HasCanonicalisationPatternsTrait):
    @classmethod
    def get_canonicalization_patterns(cls) -> tuple[RewritePattern, ...]:
        from xdsl.transforms.canonicalization_patterns.riscv import (
            ScfgwOpUsingImmediate,
        )

        return (ScfgwOpUsingImmediate(),)


@irdl_op_definition
class ScfgwOp(RdRsRsOperation[IntRegisterType, IntRegisterType, IntRegisterType]):
    """
    Write the value in rs1 to the Snitch stream configuration
    location pointed by rs2 in the memory-mapped address space.
    Register rd is always fixed to zero.

    This is a RISC-V ISA extension, part of the `Xssr' extension.
    https://pulp-platform.github.io/snitch/rm/custom_instructions/
    """

    name = "riscv_snitch.scfgw"

    traits = frozenset((ScfgwOpHasCanonicalizationPatternsTrait(),))

    def assembly_line_args(self) -> tuple[AssemblyInstructionArg, ...]:
        # rd is always zero, so we omit it when printing assembly
        return self.rs1, self.rs2

    def verify_(self) -> None:
        if cast(IntRegisterType, self.rd.type) != Registers.ZERO:
            raise VerifyException(f"scfgw rd must be ZERO, got {self.rd.type}")


@irdl_op_definition
class ScfgwiOp(RdRsImmIntegerOperation):
    """
    Write the value in rs to the Snitch stream configuration location pointed by
    immediate value in the memory-mapped address space.

    This is a RISC-V ISA extension, part of the `Xssr' extension.
    https://pulp-platform.github.io/snitch/rm/custom_instructions/
    """

    name = "riscv_snitch.scfgwi"

    def assembly_line_args(self) -> tuple[AssemblyInstructionArg, ...]:
        # rd is always zero, so we omit it when printing assembly
        return self.rs1, self.immediate

    def verify_(self) -> None:
        if cast(IntRegisterType, self.rd.type) != Registers.ZERO:
            raise VerifyException(f"scfgwi rd must be ZERO, got {self.rd.type}")


@irdl_op_definition
class FrepYieldOp(AbstractYieldOperation[Attribute], RISCVOp):
    name = "riscv_snitch.frep_yield"

    traits = traits_def(
        lambda: frozenset([IsTerminator(), HasParent(FrepInner, FrepOuter)])
    )

    def assembly_line(self) -> str | None:
        return None


@irdl_op_definition
class ReadOp(stream.ReadOperation, RISCVOp):
    name = "riscv_snitch.read"

    def assembly_line(self) -> str | None:
        return None


@irdl_op_definition
class WriteOp(stream.WriteOperation, RISCVOp):
    name = "riscv_snitch.write"

    def assembly_line(self) -> str | None:
        return None


ALLOWED_FREP_OP_TYPES = (
    FrepYieldOp,
    ReadOp,
    WriteOp,
    UnrealizedConversionCastOp,
)


class FRepOperation(IRDLOperation, RISCVInstruction):
    """
    From the Snitch paper: https://arxiv.org/abs/2002.10143

    The frep instruction marks the beginning of a floating-point kernel which should be
    repeated. It indicates how many subsequent instructions are stored in the sequence
    buffer, how often and how (operand staggering, repetition mode) each instruction is
    going to be repeated.
    """

    max_rep = operand_def(IntRegisterType)
    """Number of times to repeat the instructions."""
    body = region_def("single_block")
    """
    Instructions to repeat, containing maximum 15 instructions, with no side effects.
    """
    iter_args = var_operand_def(riscv.RISCVRegisterType)
    """
    Loop-carried variable initial values.
    """
    stagger_mask = attr_def(IntAttr)
    """
    4 bits for each operand (rs1 rs2 rs3 rd). If the bit is set, the corresponding operand
    is staggered.
    """
    stagger_count = attr_def(IntAttr)
    """
    3 bits, indicating for how many iterations the stagger should increment before it
    wraps again (up to 23 = 8).
    """
    res = var_result_def(riscv.RISCVRegisterType)
    """
    Loop-carried variable initial values.
    """

    traits = traits_def(
        lambda: frozenset((SingleBlockImplicitTerminator(FrepYieldOp),))
    )

    def __init__(
        self,
        max_rep: SSAValue | Operation,
        body: Sequence[Operation] | Sequence[Block] | Region,
        iter_args: Sequence[SSAValue | Operation] = (),
        stagger_mask: IntAttr | None = None,
        stagger_count: IntAttr | None = None,
    ):
        if stagger_mask is None:
            stagger_mask = IntAttr(0)
        if stagger_count is None:
            stagger_count = IntAttr(0)
        super().__init__(
            operands=(max_rep, iter_args),
            result_types=[[SSAValue.get(a).type for a in iter_args]],
            regions=(body,),
            attributes={
                "stagger_mask": stagger_mask,
                "stagger_count": stagger_count,
            },
        )

    @property
    def max_inst(self) -> int:
        """
        Number of instructions to be repeated.
        """
        return len([op for op in self.body.ops if isinstance(op, RISCVInstruction)])

    def assembly_line_args(self) -> tuple[AssemblyInstructionArg | None, ...]:
        return (
            self.max_rep,
            self.max_inst,
            self.stagger_mask.data,
            self.stagger_count.data,
        )

    @classmethod
    def parse(cls, parser: Parser) -> Self:
        max_rep = parser.parse_operand()
        if parser.parse_optional_punctuation(","):
            stagger_mask = parser.parse_integer(False, False)
            parser.parse_punctuation(",")
            stagger_count = parser.parse_integer(False, False)
        else:
            stagger_mask = 0
            stagger_count = 0

        remaining_attributes = parser.parse_optional_attr_dict_with_keyword()

        # Parse iteration arguments
        pos = parser.pos
        unresolved_iter_args: list[Parser.UnresolvedArgument] = []
        iter_arg_unresolved_operands: list[UnresolvedOperand] = []
        iter_arg_types: list[Attribute] = []
        if parser.parse_optional_characters("iter_args"):
            for iter_arg, iter_arg_operand in parser.parse_comma_separated_list(
                Parser.Delimiter.PAREN, lambda: parse_assignment(parser)
            ):
                unresolved_iter_args.append(iter_arg)
                iter_arg_unresolved_operands.append(iter_arg_operand)
            parser.parse_characters("->")
            iter_arg_types = parser.parse_comma_separated_list(
                Parser.Delimiter.PAREN, parser.parse_attribute
            )

        iter_arg_operands = parser.resolve_operands(
            iter_arg_unresolved_operands, iter_arg_types, pos
        )

        # Set block argument types
        iter_args = [
            u_arg.resolve(t) for u_arg, t in zip(unresolved_iter_args, iter_arg_types)
        ]

        body = parser.parse_region(iter_args)

        frep = cls(
            max_rep,
            body,
            iter_arg_operands,
            IntAttr(stagger_mask),
            IntAttr(stagger_count),
        )
        if remaining_attributes is not None:
            frep.attributes |= remaining_attributes.data

        for trait in frep.get_traits_of_type(SingleBlockImplicitTerminator):
            ensure_terminator(frep, trait)

        return frep

    def print(self, printer: Printer) -> None:
        printer.print_string(" ")
        printer.print_ssa_value(self.max_rep)
        if self.stagger_count.data and self.stagger_mask.data:
            printer.print_string(", ")
            printer.print(self.stagger_count.data)
            printer.print_string(", ")
            printer.print(self.stagger_mask.data)

        printer.print_op_attributes(
            self.attributes, reserved_attr_names=("stagger_count", "stagger_mask")
        )
        printer.print_string(" ")

        block = self.body.block

        yield_op = block.last_op
        print_block_terminators = not isinstance(yield_op, FrepYieldOp) or bool(
            yield_op.operands
        )

        if iter_args := block.args:
            printer.print_string("iter_args(")
            printer.print_list(
                zip(iter_args, self.iter_args),
                lambda pair: print_assignment(printer, *pair),
            )
            printer.print_string(") -> (")
            printer.print_list((a.type for a in iter_args), printer.print_attribute)
            printer.print_string(") ")

        printer.print_region(
            self.body,
            print_entry_block_args=False,
            print_block_terminators=print_block_terminators,
        )

    def verify_(self) -> None:
        if self.stagger_count.data:
            raise VerifyException("Non-zero stagger count currently unsupported")
        if self.stagger_mask.data:
            raise VerifyException("Non-zero stagger mask currently unsupported")
        for instruction in self.body.ops:
            if not instruction.has_trait(Pure) and not isinstance(
                instruction, ALLOWED_FREP_OP_TYPES
            ):
                raise VerifyException(
                    "Frep operation body may not contain instructions "
                    f"with side-effects, found {instruction.name}"
                )
        if len(self.iter_args) != len(self.body.block.args):
            raise VerifyException(
                f"Wrong number of block arguments, expected {len(self.iter_args)}, got "
                f"{len(self.body.block.args)}. The body must have the induction "
                f"variable and loop-carried variables as arguments."
            )
        for idx, (arg, block_arg) in enumerate(
            zip(self.iter_args, self.body.block.args)
        ):
            if block_arg.type != arg.type:
                raise VerifyException(
                    f"Block argument {idx} has wrong type, expected {arg.type}, "
                    f"got {block_arg.type}. Arguments after the "
                    f"induction variable must match the carried variables."
                )
        if len(self.body.ops) > 0 and isinstance(
            yieldop := self.body.block.last_op, FrepYieldOp
        ):
            if len(yieldop.arguments) != len(self.iter_args):
                raise VerifyException(
                    f"Expected {len(self.iter_args)} args, got {len(yieldop.arguments)}. "
                    f"The riscv_scf.frep must yield its carried variables."
                )
            for iter_arg, yield_arg in zip(self.iter_args, yieldop.arguments):
                if iter_arg.type != yield_arg.type:
                    raise VerifyException(
                        f"Expected {iter_arg.type}, got {yield_arg.type}. The "
                        f"riscv_snitch.frep's riscv_snitch.frep_yield must match carried"
                        f"variables types."
                    )


@irdl_op_definition
class FrepOuter(FRepOperation):
    """
    Repeats the instruction in the body as if the body were the body of a for loop, for
    example:

    ```
    # Repeat 4 times, stagger 1, period 2
    li a0, 4
    frep.o a0, 2, 1, 0b1010
    fadd.d fa0, ft0, ft2
    fmul.d fa0, ft3, fa0
    ```

    is equivalent to:
    ```
    fadd.d fa0, ft0, ft2
    fmul.d fa0, ft3, fa0
    fadd.d fa1, ft0, ft3
    fmul.d fa1, ft3, fa1
    fadd.d fa0, ft0, ft2
    fmul.d fa0, ft3, fa0
    fadd.d fa1, ft0, ft3
    fmul.d fa1, ft3, fa1
    ```
    """

    name = "riscv_snitch.frep_outer"

    def assembly_instruction_name(self) -> str:
        return "frep.o"


@irdl_op_definition
class FrepInner(FRepOperation):
    """
    Repeats the instruction in the body, as if each were in its own body of a for loop,
    for example:

    ```
    # Repeat three times, stagger 2, period 2
    li a0, 3
    frep.i a0, 2, 2, 0b0100
    fadd.d fa0, ft0, ft2
    fmul.d fa0, ft3, fa0
    ```

    is equivalent to:
    ```
    fadd.d fa0, ft0, ft2
    fadd.d fa0, ft1, ft3
    fadd.d fa0, ft2, ft3
    fmul.d fa0, ft3, fa0
    fmul.d fa0, ft4, fa0
    fmul.d fa0, ft5, fa0
    ```
    """

    name = "riscv_snitch.frep_inner"

    def assembly_instruction_name(self) -> str:
        return "frep.i"


@irdl_op_definition
class GetStreamOp(IRDLOperation, RISCVOp):
    name = "riscv_snitch.get_stream"

    stream = result_def(stream.StreamType[riscv.FloatRegisterType])

    def __init__(self, result_type: Attribute):
        super().__init__(result_types=[result_type])

    @classmethod
    def parse(cls, parser: Parser) -> GetStreamOp:
        parser.parse_punctuation(":")
        result_type = parser.parse_attribute()
        return GetStreamOp(result_type)

    def print(self, printer: Printer):
        printer.print_string(" : ")
        printer.print_attribute(self.stream.type)

    def assembly_line(self) -> str | None:
        return None


# endregion

# region XDMA extensions
# Documentation for these operations:
# https://pulp-platform.github.io/snitch_cluster/rm/custom_instructions.html


@irdl_op_definition
class DMSourceOp(IRDLOperation, RISCVInstruction):
    name = "riscv_snitch.dmsrc"

    ptrlo = operand_def(riscv.IntRegisterType)
    ptrhi = operand_def(riscv.IntRegisterType)

    def __init__(self, ptrlo: SSAValue | Operation, ptrhi: SSAValue | Operation):
        super().__init__(operands=[ptrlo, ptrhi])

    def assembly_line_args(self) -> tuple[AssemblyInstructionArg | None, ...]:
        return self.ptrlo, self.ptrhi


@irdl_op_definition
class DMDestinationOp(IRDLOperation, RISCVInstruction):
    name = "riscv_snitch.dmdst"

    ptrlo = operand_def(riscv.IntRegisterType)
    ptrhi = operand_def(riscv.IntRegisterType)

    def __init__(self, ptrlo: SSAValue | Operation, ptrhi: SSAValue | Operation):
        super().__init__(operands=[ptrlo, ptrhi])

    def assembly_line_args(self) -> tuple[AssemblyInstructionArg | None, ...]:
        return self.ptrlo, self.ptrhi


@irdl_op_definition
class DMStrideOp(IRDLOperation, RISCVInstruction):
    name = "riscv_snitch.dmstr"

    srcstrd = operand_def(riscv.IntRegisterType)
    dststrd = operand_def(riscv.IntRegisterType)

    def __init__(self, srcstrd: SSAValue | Operation, dststrd: SSAValue | Operation):
        super().__init__(operands=[srcstrd, dststrd])

    def assembly_line_args(self) -> tuple[AssemblyInstructionArg | None, ...]:
        return self.srcstrd, self.dststrd


@irdl_op_definition
class DMRepOp(IRDLOperation, RISCVInstruction):
    name = "riscv_snitch.dmrep"

    reps = operand_def(riscv.IntRegisterType)

    def __init__(self, reps: SSAValue | Operation):
        super().__init__(operands=[reps])

    def assembly_line_args(self) -> tuple[AssemblyInstructionArg | None, ...]:
        return (self.reps,)


@irdl_op_definition
class DMCopyOp(IRDLOperation, RISCVInstruction):
    name = "riscv_snitch.dmcpy"

    dest = result_def(riscv.IntRegisterType)
    size = operand_def(riscv.IntRegisterType)
    config = operand_def(riscv.IntRegisterType)

    def __init__(
        self,
        size: SSAValue | Operation,
        config: SSAValue | Operation,
        result_type: IntRegisterType = IntRegisterType.unallocated(),
    ):
        super().__init__(operands=[size, config], result_types=[result_type])

    def assembly_line_args(self) -> tuple[AssemblyInstructionArg | None, ...]:
        return self.dest, self.size, self.config


@irdl_op_definition
class DMStatOp(IRDLOperation, RISCVInstruction):
    name = "riscv_snitch.dmstat"

    dest = result_def(riscv.IntRegisterType)
    status = operand_def(riscv.IntRegisterType)

    def __init__(
        self,
        status: SSAValue | Operation,
        result_type: IntRegisterType = IntRegisterType.unallocated(),
    ):
        super().__init__(operands=[status], result_types=[result_type])

    def assembly_line_args(self) -> tuple[AssemblyInstructionArg | None, ...]:
        return self.dest, self.status


@irdl_op_definition
class DMCopyImmOp(IRDLOperation, RISCVInstruction):
    name = "riscv_snitch.dmcpyi"

    dest = result_def(riscv.IntRegisterType)
    size = operand_def(riscv.IntRegisterType)
    config = prop_def(UImm5Attr)

    def __init__(
        self,
        size: SSAValue | Operation,
        config: int | UImm5Attr,
        result_type: IntRegisterType = IntRegisterType.unallocated(),
    ):
        if isinstance(config, int):
            config = IntegerAttr(config, IntegerType(5, signedness=Signedness.UNSIGNED))
        super().__init__(
            operands=[size],
            properties={"config": config},
            result_types=[result_type],
        )

    def assembly_line_args(self) -> tuple[AssemblyInstructionArg | None, ...]:
        return self.dest, self.size, self.config

    def print(self, printer: Printer) -> None:
        printer.print(" ")
        printer.print_operand(self.size)
        printer.print_string(", ")
        printer.print(self.config.value.data)
        if self.attributes:
            printer.print(" ")
            printer.print_attr_dict(self.attributes)
        printer.print(" : ")
        printer.print_operation_type(self)

    @classmethod
    def parse(cls, parser: Parser) -> Self:
        size = parser.parse_operand()
        parser.parse_punctuation(",")
        config = parser.parse_integer()
        attrs = parser.parse_optional_attr_dict()
        parser.parse_punctuation(":")
        signature = parser.parse_function_type()
        result_type, *_ = signature.outputs
        op = cls(size, config, cast(IntRegisterType, result_type))
        if attrs:
            op.attributes.update(attrs)
        return op


@irdl_op_definition
class DMStatImmOp(IRDLOperation, RISCVInstruction):
    name = "riscv_snitch.dmstati"

    dest = result_def(riscv.IntRegisterType)
    status = prop_def(UImm5Attr)

    def __init__(
        self,
        status: int | UImm5Attr,
        result_type: IntRegisterType = IntRegisterType.unallocated(),
    ):
        if isinstance(status, int):
            status = IntegerAttr(status, IntegerType(5, signedness=Signedness.UNSIGNED))
        super().__init__(
            properties={"status": status},
            result_types=[result_type],
        )

    def assembly_line_args(self) -> tuple[AssemblyInstructionArg | None, ...]:
        return self.dest, self.status

    def print(self, printer: Printer) -> None:
        printer.print(" ")
        printer.print(self.status.value.data)
        if self.attributes:
            printer.print(" ")
            printer.print_attr_dict(self.attributes)
        printer.print(" : ")
        printer.print_operation_type(self)

    @classmethod
    def parse(cls, parser: Parser) -> Self:
        status = parser.parse_integer()
        attrs = parser.parse_optional_attr_dict()
        parser.parse_punctuation(":")
        signature = parser.parse_function_type()
        result_type, *_ = signature.outputs
        op = cls(status, cast(IntRegisterType, result_type))
        if attrs:
            op.attributes.update(attrs)
        return op


# endregion

RISCV_Snitch = Dialect(
    "riscv_snitch",
    [
        ScfgwOp,
        ScfgwiOp,
        FrepOuter,
        FrepInner,
        FrepYieldOp,
        ReadOp,
        WriteOp,
        GetStreamOp,
        DMSourceOp,
        DMDestinationOp,
        DMStrideOp,
        DMRepOp,
        DMCopyOp,
        DMCopyImmOp,
        DMStatOp,
        DMStatImmOp,
    ],
    [],
)



from collections.abc import Callable, Sequence
from typing import TypeGuard, cast

from attr import dataclass

from xdsl.dialects.arith import Addf, Divf, FloatingPointLikeBinaryOp, Mulf, Subf
from xdsl.dialects.builtin import (
    AnyFloat,
    ContainerType,
    IntAttr,
    MemRefType,
    ModuleOp,
    ShapedType,
    TensorType,
)
from xdsl.dialects.func import FuncOp
from xdsl.dialects.linalg import FillOp
from xdsl.dialects.stencil import (
    AccessOp,
    ApplyOp,
    ExternalLoadOp,
    FieldType,
    IndexAttr,
    LoadOp,
    ReturnOp,
    StencilBoundsAttr,
    StoreOp,
    TempType,
)
from xdsl.dialects.tensor import EmptyOp, ExtractSliceOp
from xdsl.ir import (
    Attribute,
    MLContext,
    Operation,
)
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    GreedyRewritePatternApplier,
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    TypeConversionPattern,
    attr_type_rewrite_pattern,
    op_type_rewrite_pattern,
)
from xdsl.rewriter import InsertPoint
from xdsl.utils.hints import isa


def get_required_result_type(op: Operation) -> TensorType[Attribute] | None:
    for result in op.results:
        for use in result.uses:
            if (
                isinstance(use.operation, ReturnOp)
                and (p_op := use.operation.parent_op()) is not None
            ):
                for ret in p_op.results:
                    if is_tensorized(ret.type):
                        if isa(ret.type, TempType[Attribute]) and isa(
                            r_type := ret.type.get_element_type(), TensorType[Attribute]
                        ):
                            return r_type
                # abort when encountering an un-tensorized ReturnOp successor
                return None
            for ret in use.operation.results:
                if isa(r_type := ret.type, TensorType[Attribute]):
                    return r_type


def needs_update_shape(
    op_type: Attribute, succ_req_type: TensorType[Attribute]
) -> bool:
    assert isa(op_type, TensorType[Attribute])
    return op_type.get_shape() != succ_req_type.get_shape()


def stencil_field_to_tensor(field: FieldType[Attribute]) -> FieldType[Attribute]:
    if field.get_num_dims() != 3:
        return field
    typ = TensorType(field.get_element_type(), [field.get_shape()[-1]])
    assert isinstance(field.bounds, StencilBoundsAttr)
    assert isinstance(field.bounds.lb, IndexAttr)
    assert isinstance(field.bounds.ub, IndexAttr)
    bounds = list(zip(field.bounds.lb, field.bounds.ub))[:-1]
    return FieldType[Attribute](bounds, typ)


def stencil_temp_to_tensor(field: TempType[Attribute]) -> TempType[Attribute]:
    if field.get_num_dims() != 3:
        return field
    typ = TensorType(field.get_element_type(), [field.get_shape()[-1]])
    assert isinstance(field.bounds, StencilBoundsAttr)
    assert isinstance(field.bounds.lb, IndexAttr)
    assert isinstance(field.bounds.ub, IndexAttr)
    bounds = list(zip(field.bounds.lb, field.bounds.ub))[:-1]
    return TempType[Attribute](bounds, typ)


def stencil_memref_to_tensor(field: MemRefType[Attribute]) -> MemRefType[Attribute]:
    if field.get_num_dims() != 3:
        return field
    typ = TensorType(field.get_element_type(), [field.get_shape()[-1]])
    return MemRefType[Attribute](typ, field.get_shape()[:-1])


class StencilFieldConversion(TypeConversionPattern):
    @attr_type_rewrite_pattern
    def convert_type(self, typ: FieldType[Attribute]) -> FieldType[Attribute]:
        return stencil_field_to_tensor(typ)


class StencilTempConversion(TypeConversionPattern):
    @attr_type_rewrite_pattern
    def convert_type(self, typ: TempType[Attribute]) -> TempType[Attribute]:
        return stencil_temp_to_tensor(typ)


class StencilMemRefConversion(TypeConversionPattern):
    @attr_type_rewrite_pattern
    def convert_type(self, typ: MemRefType[Attribute]) -> MemRefType[Attribute]:
        return stencil_memref_to_tensor(typ)


@dataclass(frozen=True)
class AccessOpTensorize(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: AccessOp, rewriter: PatternRewriter, /):
        if not is_tensorized(op.operands[0].type) or len(op.offset) != 3:
            return
        xy_offsets, z_offset = (
            tuple(o for o in op.offset)[:-1],
            tuple(o for o in op.offset)[-1],
        )
        a = AccessOp.get(op.temp, xy_offsets, op.offset_mapping)
        # this conditional controls if ExtractSliceOps for x/y accesses should be generated
        # if xy_offsets[0] != 0 or xy_offsets[1] != 0:
        #     rewriter.replace_matched_op(a)
        #     return
        assert isa(op.temp.type, TempType[Attribute])
        assert is_tensor(element_t := op.temp.type.get_element_type())
        extract = ExtractSliceOp.from_static_parameters(
            a, [z_offset], element_t.get_shape()
        )
        rewriter.insert_op(a, InsertPoint.before(op))
        rewriter.replace_matched_op(extract)


def arithBinaryOpTensorize(
    type_constructor: Callable[..., FloatingPointLikeBinaryOp],
    op: FloatingPointLikeBinaryOp,
    rewriter: PatternRewriter,
    /,
):
    if is_tensor(op.result.type):
        return
    if is_tensor(op.lhs.type) and is_tensor(op.rhs.type):
        rewriter.replace_matched_op(
            type_constructor(op.lhs, op.rhs, flags=None, result_type=op.lhs.type)
        )
    elif is_tensor(op.lhs.type) and is_scalar(op.rhs.type):
        emptyop = EmptyOp((), op.lhs.type)
        fillop = FillOp((op.rhs,), (emptyop,), (op.lhs.type,))
        rewriter.insert_op(emptyop, InsertPoint.before(op))
        rewriter.insert_op(fillop, InsertPoint.before(op))
        rewriter.replace_matched_op(
            type_constructor(op.lhs, fillop, flags=None, result_type=op.lhs.type)
        )
    elif is_scalar(op.lhs.type) and is_tensor(op.rhs.type):
        emptyop = EmptyOp((), op.rhs.type)
        fillop = FillOp((op.lhs,), (emptyop,), (op.rhs.type,))
        rewriter.insert_op(emptyop, InsertPoint.before(op))
        rewriter.insert_op(fillop, InsertPoint.before(op))
        rewriter.replace_matched_op(
            type_constructor(fillop, op.rhs, flags=None, result_type=op.rhs.type)
        )


class ArithAddfOpTensorize(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: Addf, rewriter: PatternRewriter, /):
        arithBinaryOpTensorize(Addf, op, rewriter)


class ArithSubfOpTensorize(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: Subf, rewriter: PatternRewriter, /):
        arithBinaryOpTensorize(Subf, op, rewriter)


class ArithMulfOpTensorize(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: Mulf, rewriter: PatternRewriter, /):
        arithBinaryOpTensorize(Mulf, op, rewriter)


class ArithDivfOpTensorize(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: Divf, rewriter: PatternRewriter, /):
        arithBinaryOpTensorize(Divf, op, rewriter)


@dataclass(frozen=True)
class ApplyOpTensorize(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: ApplyOp, rewriter: PatternRewriter, /):
        if all(is_tensorized(arg.type) for arg in op.args) and all(
            not is_tensorized(r.type) for r in op.res
        ):
            b = op.region.block
            # TODO check if block args need updating
            for _ in range(len(b.args)):
                assert isa(arg_type := b.args[0].type, TempType[Attribute])
                arg = b.insert_arg(stencil_temp_to_tensor(arg_type), len(b.args))
                b.args[0].replace_by(arg)
                b.erase_arg(b.args[0], safe_erase=True)
            rewriter.replace_matched_op(
                ApplyOp.get(
                    op.args,
                    op.region.clone(),
                    [
                        stencil_temp_to_tensor(cast(TempType[Attribute], r.type))
                        for r in op.res
                    ],
                )
            )


class FuncOpTensorize(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: FuncOp, rewriter: PatternRewriter, /):
        for arg in op.args:
            assert isa(arg.type, MemRefType[Attribute])
            op.replace_argument_type(arg, stencil_memref_to_tensor(arg.type))


def is_tensorized(
    typ: Attribute,
):
    assert isinstance(typ, ShapedType)
    assert isinstance(typ, ContainerType)
    return len(typ.get_shape()) == 2 and isinstance(typ.get_element_type(), TensorType)


def is_tensor(typ: Attribute) -> TypeGuard[TensorType[Attribute]]:
    return isinstance(typ, TensorType)


def is_scalar(typ: Attribute) -> TypeGuard[AnyFloat]:
    return isinstance(typ, AnyFloat)


class ExternalLoadOpTensorize(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: ExternalLoadOp, rewriter: PatternRewriter, /):
        if is_tensorized(op.field.type) and not is_tensorized(op.result.type):
            assert isa(op.result.type, FieldType[Attribute])
            rewriter.replace_matched_op(
                ExternalLoadOp.get(op.field, stencil_field_to_tensor(op.result.type))
            )


class LoadOpTensorize(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: LoadOp, rewriter: PatternRewriter, /):
        if is_tensorized(op.field.type) and not is_tensorized(op.res.type):
            assert isa(op.res.type, TempType[Attribute])
            assert isinstance(bounds := op.res.type.bounds, StencilBoundsAttr)
            rewriter.replace_matched_op(
                LoadOp.get(
                    op.field,
                    IndexAttr.get(*[lb for lb in bounds.lb][:-1]),
                    IndexAttr.get(*[ub for ub in bounds.ub][:-1]),
                )
            )


class StoreOpTensorize(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: StoreOp, rewriter: PatternRewriter, /):
        if (
            is_tensorized(op.field.type)
            and isinstance(op.field.type, ShapedType)
            and len(op.bounds.lb) != len(op.field.type.get_shape())
        ):
            rewriter.replace_matched_op(
                StoreOp.get(
                    op.temp,
                    op.field,
                    StencilBoundsAttr(
                        zip(list(op.bounds.lb.array)[:-1], list(op.bounds.ub)[:-1])
                    ),
                )
            )


class AccessOpUpdateShape(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: AccessOp, rewriter: PatternRewriter, /):
        if typ := get_required_result_type(op):
            if needs_update_shape(op.res.type, typ):
                rewriter.replace_matched_op(
                    AccessOp.build(
                        operands=[op.temp], attributes=op.attributes, result_types=[typ]
                    )
                )


class ExtractSliceOpUpdateShape(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: ExtractSliceOp, rewriter: PatternRewriter, /):
        if typ := get_required_result_type(op):
            if needs_update_shape(op.result.type, typ):
                # offsets = op.static_offsets.data.data
                if isa(offsets := op.static_offsets.data.data, Sequence[IntAttr]):
                    new_offsets = [o.data for o in offsets]
                else:
                    assert isa(offsets, Sequence[int])
                    new_offsets = offsets
                rewriter.replace_matched_op(
                    ExtractSliceOp.from_static_parameters(
                        op.source, new_offsets, typ.get_shape()
                    )
                )


def arithBinaryOpUpdateShape(
    type_constructor: Callable[..., FloatingPointLikeBinaryOp],
    op: FloatingPointLikeBinaryOp,
    rewriter: PatternRewriter,
    /,
):
    if typ := get_required_result_type(op):
        if needs_update_shape(op.result.type, typ):
            rewriter.replace_matched_op(
                type_constructor(op.lhs, op.rhs, flags=None, result_type=typ)
            )


class ArithAddfOpUpdateShape(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: Addf, rewriter: PatternRewriter, /):
        arithBinaryOpUpdateShape(Addf, op, rewriter)


class ArithSubfOpUpdateShape(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: Subf, rewriter: PatternRewriter, /):
        arithBinaryOpUpdateShape(Subf, op, rewriter)


class ArithMulfOpUpdateShape(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: Mulf, rewriter: PatternRewriter, /):
        arithBinaryOpUpdateShape(Mulf, op, rewriter)


class ArithDivfOpUpdateShape(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: Divf, rewriter: PatternRewriter, /):
        arithBinaryOpUpdateShape(Divf, op, rewriter)


class EmptyOpUpdateShape(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: EmptyOp, rewriter: PatternRewriter, /):
        if typ := get_required_result_type(op):
            if needs_update_shape(op.results[0].type, typ):
                rewriter.replace_matched_op(EmptyOp((), typ))


class FillOpUpdateShape(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: FillOp, rewriter: PatternRewriter, /):
        if typ := get_required_result_type(op):
            if needs_update_shape(op.results[0].type, typ):
                rewriter.replace_matched_op(
                    FillOp(op.inputs, op.outputs, [typ] * len(op.outputs))
                )


@dataclass(frozen=True)
class StencilTensorizeZDimension(ModulePass):
    name = "stencil-tensorize-z-dimension"

    def apply(self, ctx: MLContext, op: ModuleOp) -> None:
        module_pass = PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    FuncOpTensorize(),
                    ExternalLoadOpTensorize(),
                    LoadOpTensorize(),
                    ApplyOpTensorize(),
                    StoreOpTensorize(),
                    # AccessOpTensorize(),   # this doesn't work here, using second pass
                ]
            ),
            walk_reverse=False,
            apply_recursively=False,
        )
        module_pass.rewrite_module(op)
        stencil_pass = PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    AccessOpTensorize(),
                    ArithAddfOpTensorize(),
                    ArithMulfOpTensorize(),
                    ArithSubfOpTensorize(),
                    ArithDivfOpTensorize(),
                ]
            ),
            walk_reverse=False,
            apply_recursively=False,
        )
        stencil_pass.rewrite_module(op)
        backpropagate_stencil_shapes = PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    # AccessOpUpdateShape(),
                    ExtractSliceOpUpdateShape(),
                    EmptyOpUpdateShape(),
                    FillOpUpdateShape(),
                    ArithAddfOpUpdateShape(),
                    ArithSubfOpUpdateShape(),
                    ArithMulfOpUpdateShape(),
                    ArithDivfOpUpdateShape(),
                ]
            ),
            walk_reverse=True,
            apply_recursively=False,
        )
        backpropagate_stencil_shapes.rewrite_module(op)



from __future__ import annotations

import inspect
from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable, Sequence
from dataclasses import dataclass, field
from functools import wraps
from types import UnionType
from typing import TypeVar, Union, final, get_args, get_origin

from typing_extensions import deprecated

from xdsl.builder import BuilderListener
from xdsl.dialects.builtin import ArrayAttr, DictionaryAttr, ModuleOp
from xdsl.ir import (
    Attribute,
    Block,
    BlockArgument,
    ErasedSSAValue,
    Operation,
    ParametrizedAttribute,
    Region,
    SSAValue,
)
from xdsl.rewriter import InsertPoint, Rewriter
from xdsl.utils.hints import isa


@dataclass(eq=False)
class PatternRewriterListener(BuilderListener):
    """A listener for pattern rewriter events."""

    operation_removal_handler: list[Callable[[Operation], None]] = field(
        default_factory=list, kw_only=True
    )
    """Callbacks that are called when an operation is removed."""

    operation_modification_handler: list[Callable[[Operation], None]] = field(
        default_factory=list, kw_only=True
    )
    """Callbacks that are called when an operation is modified."""

    operation_replacement_handler: list[
        Callable[[Operation, Sequence[SSAValue | None]], None]
    ] = field(default_factory=list, kw_only=True)
    """Callbacks that are called when an operation is replaced."""

    def handle_operation_removal(self, op: Operation) -> None:
        """Pass the operation that will be removed to the registered callbacks."""
        for handler in self.operation_removal_handler:
            handler(op)

    def handle_operation_modification(self, op: Operation) -> None:
        """Pass the operation that was just modified to the registered callbacks."""
        for handler in self.operation_modification_handler:
            handler(op)

    def handle_operation_replacement(
        self, op: Operation, new_results: Sequence[SSAValue | None]
    ) -> None:
        """Pass the operation that will be replaced to the registered callbacks."""
        for handler in self.operation_replacement_handler:
            handler(op, new_results)

    def extend_from_listener(self, listener: BuilderListener | PatternRewriterListener):
        """Forward all callbacks from `listener` to this listener."""
        super().extend_from_listener(listener)
        if isinstance(listener, PatternRewriterListener):
            self.operation_removal_handler.extend(listener.operation_removal_handler)
            self.operation_modification_handler.extend(
                listener.operation_modification_handler
            )
            self.operation_replacement_handler.extend(
                listener.operation_replacement_handler
            )


@dataclass(eq=False)
class PatternRewriter(PatternRewriterListener):
    """
    A rewriter used during pattern matching.
    Once an operation is matched, this rewriter is used to apply
    modification to the operation and its children.
    """

    current_operation: Operation
    """The matched operation."""

    has_done_action: bool = field(default=False, init=False)
    """Has the rewriter done any action during the current match."""

    def insert_op(
        self, op: Operation | Sequence[Operation], insertion_point: InsertPoint
    ):
        """Insert operations at a certain location in a block."""
        self.has_done_action = True
        op = (op,) if isinstance(op, Operation) else op
        if not op:
            return
        Rewriter.insert_op(op, insertion_point)

        for op_ in op:
            self.handle_operation_insertion(op_)

    def insert_op_before_matched_op(self, op: Operation | Sequence[Operation]):
        """Insert operations before the matched operation."""
        self.insert_op(op, InsertPoint.before(self.current_operation))

    def insert_op_after_matched_op(self, op: Operation | Sequence[Operation]):
        """Insert operations after the matched operation."""
        self.insert_op(op, InsertPoint.after(self.current_operation))

    @deprecated("Please use `insert_op` instead")
    def insert_op_at_end(self, op: Operation | Sequence[Operation], block: Block):
        """Insert operations at the end of a block."""
        self.insert_op(op, InsertPoint.at_end(block))

    @deprecated("Please use `insert_op` instead")
    def insert_op_at_start(self, op: Operation | Sequence[Operation], block: Block):
        """Insert operations at the start of a block."""
        self.insert_op(op, InsertPoint.at_start(block))

    @deprecated("Please use `insert_op` instead")
    def insert_op_before(
        self, op: Operation | Sequence[Operation], target_op: Operation
    ):
        """Insert operations before an operation."""
        self.insert_op(op, InsertPoint.before(target_op))

    @deprecated("Please use `insert_op` instead")
    def insert_op_after(
        self, op: Operation | Sequence[Operation], target_op: Operation
    ):
        """Insert operations after an operation."""
        self.insert_op(op, InsertPoint.after(target_op))

    def erase_matched_op(self, safe_erase: bool = True):
        """
        Erase the operation that was matched to.
        If safe_erase is True, check that the operation has no uses.
        Otherwise, replace its uses with ErasedSSAValue.
        """
        self.erase_op(self.current_operation, safe_erase=safe_erase)

    def erase_op(self, op: Operation, safe_erase: bool = True):
        """
        Erase an operation.
        If safe_erase is True, check that the operation has no uses.
        Otherwise, replace its uses with ErasedSSAValue.
        """
        self.has_done_action = True
        self.handle_operation_removal(op)
        Rewriter.erase_op(op, safe_erase=safe_erase)

    def _replace_all_uses_with(
        self, from_: SSAValue, to: SSAValue | None, safe_erase: bool = True
    ):
        """Replace all uses of an SSA value with another SSA value."""
        for use in from_.uses:
            self.handle_operation_modification(use.operation)
        if to is None:
            from_.erase(safe_erase=safe_erase)
        else:
            from_.replace_by(to)

    def replace_matched_op(
        self,
        new_ops: Operation | Sequence[Operation],
        new_results: Sequence[SSAValue | None] | None = None,
        safe_erase: bool = True,
    ):
        """
        Replace the matched operation with new operations.
        Also, optionally specify SSA values to replace the operation results.
        If safe_erase is True, check that the operation has no uses.
        Otherwise, replace its uses with ErasedSSAValue.
        """
        self.replace_op(
            self.current_operation, new_ops, new_results, safe_erase=safe_erase
        )

    def replace_op(
        self,
        op: Operation,
        new_ops: Operation | Sequence[Operation],
        new_results: Sequence[SSAValue | None] | None = None,
        safe_erase: bool = True,
    ):
        """
        Replace an operation with new operations.
        Also, optionally specify SSA values to replace the operation results.
        If safe_erase is True, check that the operation has no uses.
        Otherwise, replace its uses with ErasedSSAValue.
        """
        self.has_done_action = True
        if isinstance(new_ops, Operation):
            new_ops = [new_ops]

        # First, insert the new operations before the matched operation
        self.insert_op(new_ops, InsertPoint.before(op))

        if isinstance(new_ops, Operation):
            new_ops = [new_ops]
        if new_results is None:
            new_results = [] if len(new_ops) == 0 else new_ops[-1].results

        if len(op.results) != len(new_results):
            raise ValueError(
                f"Expected {len(op.results)} new results, but got {len(new_results)}"
            )

        # Then, replace the results with new ones
        self.handle_operation_replacement(op, new_results)
        for old_result, new_result in zip(op.results, new_results):
            self._replace_all_uses_with(old_result, new_result)

            # Preserve name hints for ops with multiple results
            if new_result is not None and not new_result.name_hint:
                new_result.name_hint = old_result.name_hint

        # Add name hints for existing ops, only if there is a single new result
        if (
            len(new_results) == 1
            and (only_result := new_results[0]) is not None
            and (name_hint := only_result.name_hint) is not None
        ):
            for new_op in new_ops:
                for res in new_op.results:
                    if not res.name_hint:
                        res.name_hint = name_hint

        # Then, erase the original operation
        self.erase_op(op, safe_erase=safe_erase)

    def modify_block_argument_type(self, arg: BlockArgument, new_type: Attribute):
        """Modify the type of a block argument."""
        self.has_done_action = True
        arg.type = new_type

        for use in arg.uses:
            self.handle_operation_modification(use.operation)

    def insert_block_argument(
        self, block: Block, index: int, arg_type: Attribute
    ) -> BlockArgument:
        """Insert a new block argument."""
        self.has_done_action = True
        return block.insert_arg(arg_type, index)

    def erase_block_argument(self, arg: BlockArgument, safe_erase: bool = True) -> None:
        """
        Erase a new block argument.
        If safe_erase is true, then raise an exception if the block argument has still
        uses, otherwise, replace it with an ErasedSSAValue.
        """
        self.has_done_action = True
        self._replace_all_uses_with(arg, None, safe_erase=safe_erase)
        arg.block.erase_arg(arg, safe_erase)

    def inline_block(
        self,
        block: Block,
        insertion_point: InsertPoint,
        arg_values: Sequence[SSAValue] = (),
    ):
        """
        Move the block operations to the specified insertion point.
        """
        self.has_done_action = True
        Rewriter.inline_block(block, insertion_point, arg_values=arg_values)

    @deprecated("Please use `inline_block` instead")
    def inline_block_at_end(
        self, block: Block, target_block: Block, arg_values: Sequence[SSAValue] = ()
    ):
        """
        Move the block operations to the end of another block.
        This block should not be a parent of the block to move to.
        """
        self.inline_block(
            block, InsertPoint.at_end(target_block), arg_values=arg_values
        )

    @deprecated("Please use `inline_block` instead")
    def inline_block_at_start(
        self, block: Block, target_block: Block, arg_values: Sequence[SSAValue] = ()
    ):
        """
        Move the block operations to the start of another block.
        This block should not be a parent of the block to move to.
        """
        self.inline_block(
            block, InsertPoint.at_start(target_block), arg_values=arg_values
        )

    def inline_block_before_matched_op(
        self, block: Block, arg_values: Sequence[SSAValue] = ()
    ):
        """
        Move the block operations before the matched operation.
        The block should not be a parent of the operation.
        """
        self.inline_block(
            block, InsertPoint.before(self.current_operation), arg_values=arg_values
        )

    @deprecated("Please use `inline_block` instead")
    def inline_block_before(
        self, block: Block, op: Operation, arg_values: Sequence[SSAValue] = ()
    ):
        """
        Move the block operations before the given operation.
        The block should not be a parent of the operation.
        """
        self.inline_block(block, InsertPoint.before(op), arg_values=arg_values)

    def inline_block_after_matched_op(
        self, block: Block, arg_values: Sequence[SSAValue] = ()
    ):
        """
        Move the block operations after the matched operation.
        The block should not be a parent of the operation.
        """
        self.inline_block(
            block, InsertPoint.after(self.current_operation), arg_values=arg_values
        )

    @deprecated("Please use `inline_block` instead")
    def inline_block_after(
        self, block: Block, op: Operation, arg_values: Sequence[SSAValue] = ()
    ):
        """
        Move the block operations after the given operation.
        The block should not be a parent of the operation.
        """
        self.inline_block(block, InsertPoint.after(op), arg_values=arg_values)

    def move_region_contents_to_new_regions(self, region: Region) -> Region:
        """Move the region blocks to a new region."""
        self.has_done_action = True
        return Rewriter.move_region_contents_to_new_regions(region)

    def inline_region_before(self, region: Region, target: Block) -> None:
        """Move the region blocks to an existing region."""
        self.has_done_action = True
        Rewriter.inline_region_before(region, target)

    def inline_region_after(self, region: Region, target: Block) -> None:
        """Move the region blocks to an existing region."""
        self.has_done_action = True
        Rewriter.inline_region_after(region, target)

    def inline_region_at_start(self, region: Region, target: Region) -> None:
        """Move the region blocks to an existing region."""
        self.has_done_action = True
        Rewriter.inline_region_at_start(region, target)

    def inline_region_at_end(self, region: Region, target: Region) -> None:
        """Move the region blocks to an existing region."""
        self.has_done_action = True
        Rewriter.inline_region_at_end(region, target)


class RewritePattern(ABC):
    """
    A side-effect free rewrite pattern matching on a DAG.
    """

    # The / in the function signature makes the previous arguments positional, see
    # https://peps.python.org/pep-0570/
    # This is used by the op_type_rewrite_pattern
    @abstractmethod
    def match_and_rewrite(self, op: Operation, rewriter: PatternRewriter, /):
        """
        Match an operation, and optionally perform a rewrite using the rewriter.
        """
        ...


_RewritePatternT = TypeVar("_RewritePatternT", bound=RewritePattern)
_OperationT = TypeVar("_OperationT", bound=Operation)


def op_type_rewrite_pattern(
    func: Callable[[_RewritePatternT, _OperationT, PatternRewriter], None]
) -> Callable[[_RewritePatternT, Operation, PatternRewriter], None]:
    """
    This function is intended to be used as a decorator on a RewritePatter
    method. It uses type hints to match on a specific operation type before
    calling the decorated function.
    """
    # Get the operation argument and check that it is a subclass of Operation
    params = [param for param in inspect.signature(func).parameters.values()]
    if len(params) != 3:
        raise Exception(
            "op_type_rewrite_pattern expects the decorated function to "
            "have two non-self arguments."
        )
    is_method = params[0].name == "self"
    if is_method:
        if len(params) != 3:
            raise Exception(
                "op_type_rewrite_pattern expects the decorated method to "
                "have two non-self arguments."
            )
    else:
        if len(params) != 2:
            raise Exception(
                "op_type_rewrite_pattern expects the decorated function to "
                "have two arguments."
            )
    expected_type: type[_OperationT] = params[-2].annotation

    expected_types = (expected_type,)
    if get_origin(expected_type) in [Union, UnionType]:
        expected_types = get_args(expected_type)

    if not all(issubclass(t, Operation) for t in expected_types):
        raise Exception(
            "op_type_rewrite_pattern expects the first non-self argument "
            "type hint to be an `Operation` subclass or a union of `Operation` "
            "subclasses."
        )

    def impl(self: _RewritePatternT, op: Operation, rewriter: PatternRewriter) -> None:
        if isinstance(op, expected_type):
            func(self, op, rewriter)

    return impl


@dataclass
class TypeConversionPattern(RewritePattern):
    """
    Base pattern for type conversion. It is supposed to be inherited from, then one can
    implement `convert_type` to define the conversion.

    It will convert an Operations' result types, dictionary attributes, and block arguments.

    One can use `@attr_type_rewrite_pattern` on this defined method to automatically filter
    on the Attribute type used.

    This base pattern defines two flags:

    - `recursive` (defaulting to False): recurse over structured attributes to convert
      parameters.
      e.g. a recusrive `i32` to `index` conversion will convert `vector<i32>` to
      `vector<index>`.
    - `ops` (defaulting to any Operation) is a tuple of Operation types on which to apply
      the defined attribute conversion.
    """

    recursive: bool = False
    """
    recurse over structured attributes to convert parameters.
    Defaults to False.
    """
    ops: tuple[type[Operation], ...] | None = None
    """
    A tuple of Operation types on which to apply the defined attribute conversion.
    Defaults to any operation type.
    """

    @abstractmethod
    def convert_type(self, typ: Attribute, /) -> Attribute | None:
        """
        The method to implement to define a TypeConversionPattern

        This defines how the input Attribute should be converted.
        It allows returning None, meaning "this attribute should not
        be converted".
        """
        raise NotImplementedError()

    @final
    def _convert_type_rec(self, typ: Attribute) -> Attribute | None:
        """
        Provided recursion over structed/parameterized Attributes.
        """
        inp = typ
        if self.recursive:
            if isinstance(typ, ParametrizedAttribute):
                parameters = list(
                    self._convert_type_rec(p) or p for p in typ.parameters
                )
                inp = type(typ).new(parameters)
            if isa(typ, ArrayAttr[Attribute]):
                parameters = tuple(self._convert_type_rec(p) or p for p in typ)
                inp = type(typ).new(parameters)
            if isa(typ, DictionaryAttr):
                parameters = {k: self._convert_type_rec(v) for k, v in typ.data.items()}
                inp = type(typ).new(parameters)
        converted = self.convert_type(inp)
        return converted if converted is not None else inp

    @final
    def match_and_rewrite(self, op: Operation, rewriter: PatternRewriter):
        """
        Pattern application implementation
        """
        if self.ops and not isinstance(op, self.ops):
            return
        new_result_types: list[Attribute] = []
        new_attributes: dict[str, Attribute] = {}
        new_properties: dict[str, Attribute] = {}
        changed: bool = False
        for result in op.results:
            converted = self._convert_type_rec(result.type)
            new_result_types.append(converted or result.type)
            if converted is not None and converted != result.type:
                changed = True
        for name, attribute in op.attributes.items():
            converted = self._convert_type_rec(attribute)
            new_attributes[name] = converted or attribute
            if converted is not None and converted != attribute:
                changed = True
        for name, attribute in op.properties.items():
            converted = self._convert_type_rec(attribute)
            new_properties[name] = converted or attribute
            if converted is not None and converted != attribute:
                changed = True
        for region in op.regions:
            for block in region.blocks:
                for arg in block.args:
                    converted = self.convert_type(arg.type)
                    if converted is not None and converted != arg.type:
                        rewriter.modify_block_argument_type(arg, converted)
        if changed:
            regions = [op.detach_region(r) for r in op.regions]
            new_op = type(op).create(
                operands=op.operands,
                result_types=new_result_types,
                properties=new_properties,
                attributes=new_attributes,
                successors=op.successors,
                regions=regions,
            )
            rewriter.replace_matched_op(new_op)
            for new, old in zip(new_op.results, op.results):
                new.name_hint = old.name_hint


_TypeConversionPatternT = TypeVar(
    "_TypeConversionPatternT", bound=TypeConversionPattern
)
_AttributeT = TypeVar("_AttributeT", bound=Attribute)
_ConvertedT = TypeVar("_ConvertedT", bound=Attribute)


def attr_type_rewrite_pattern(
    func: Callable[[_TypeConversionPatternT, _AttributeT], _ConvertedT | None]
) -> Callable[[_TypeConversionPatternT, Attribute], Attribute | None]:
    """
    This function is intended to be used as a decorator on a TypeConversionPattern
    method. It uses type hints to match on a specific attribute type before
    calling the decorated function.
    """
    params = list(inspect.signature(func).parameters.values())
    expected_type: type[_AttributeT] = params[-1].annotation

    @wraps(func)
    def impl(self: _TypeConversionPatternT, typ: Attribute) -> Attribute | None:
        if isa(typ, expected_type):
            return func(self, typ)
        return None

    return impl


@dataclass(eq=False, repr=False)
class GreedyRewritePatternApplier(RewritePattern):
    """
    Apply a list of patterns in order until one pattern matches,
    and then use this rewrite.
    """

    rewrite_patterns: list[RewritePattern]
    """The list of rewrites to apply in order."""

    def match_and_rewrite(self, op: Operation, rewriter: PatternRewriter) -> None:
        for pattern in self.rewrite_patterns:
            pattern.match_and_rewrite(op, rewriter)
            if rewriter.has_done_action:
                return
        return


@dataclass(eq=False)
class Worklist:
    _op_stack: list[Operation | None] = field(default_factory=list, init=False)
    """
    The list of operations to iterate over, used as a last-in-first-out stack.
    Operations are added and removed at the end of the list.
    Operation that are `None` are meant to be discarded, and are used to
    keep removal of operations O(1).
    """

    _map: dict[Operation, int] = field(default_factory=dict, init=False)
    """
    The map of operations to their index in the stack.
    It is used to check if an operation is already in the stack, and to
    remove it in O(1).
    """

    def is_empty(self) -> bool:
        """Check if the worklist is empty."""
        while self._op_stack and self._op_stack[-1] is None:
            self._op_stack.pop()
        return not bool(self._op_stack)

    def push(self, op: Operation):
        """
        Push an operation to the end of the worklist, if it is not already in it.
        """
        if op not in self._map:
            self._map[op] = len(self._op_stack)
            self._op_stack.append(op)

    def pop(self) -> Operation | None:
        """Pop the operation at the end of the worklist."""
        # All `None` operations at the end of the stack are discarded,
        # as they were removed previously.
        # We either return `None` if the stack is empty, or the last operation
        # that is not `None`.
        while self._op_stack:
            op = self._op_stack.pop()
            if op is not None:
                del self._map[op]
                return op
        return None

    def remove(self, op: Operation):
        """Remove an operation from the worklist."""
        if op in self._map:
            index = self._map[op]
            self._op_stack[index] = None
            del self._map[op]


@dataclass(eq=False, repr=False)
class PatternRewriteWalker:
    """
    Walks the IR in the block and instruction order, and rewrite it in place.
    Previous references to the walked operations are invalid after the walk.
    Can walk either first the regions, or first the owner operation.
    The walker will also walk recursively on the created operations.
    """

    pattern: RewritePattern
    """Pattern to apply during the walk."""

    walk_regions_first: bool = field(default=False)
    """
    Choose if the walker should first walk the operation regions first,
    or the operation itself.
    """

    apply_recursively: bool = field(default=True)
    """Apply recursively rewrites on new operations."""

    walk_reverse: bool = field(default=False)
    """
    Walk the regions and blocks in reverse order.
    That way, all uses are replaced before the definitions.
    """

    listener: PatternRewriterListener = field(default_factory=PatternRewriterListener)
    """The listener that will be called when an operation or block is modified."""

    _worklist: Worklist = field(default_factory=Worklist, init=False)
    """The worklist of operations to walk over."""

    def _add_operands_to_worklist(self, operands: Iterable[SSAValue]) -> None:
        """
        Add defining operations of SSA values to the worklist if they have only
        one use. This is a heuristic based on the fact that single-use operations
        have more canonicalization opportunities.
        """
        for operand in operands:
            if (
                len(operand.uses) == 1
                and not isinstance(operand, ErasedSSAValue)
                and isinstance((op := operand.owner), Operation)
            ):
                self._worklist.push(op)

    def _handle_operation_insertion(self, op: Operation) -> None:
        """Handle insertion of an operation."""
        if self.apply_recursively:
            self._worklist.push(op)

    def _handle_operation_removal(self, op: Operation) -> None:
        """Handle removal of an operation."""
        if self.apply_recursively:
            self._add_operands_to_worklist(op.operands)
        self._worklist.remove(op)

    def _handle_operation_modification(self, op: Operation) -> None:
        """Handle modification of an operation."""
        if self.apply_recursively:
            self._worklist.push(op)

    def _handle_operation_replacement(
        self, op: Operation, new_results: Sequence[SSAValue | None]
    ) -> None:
        """Handle replacement of an operation."""
        if self.apply_recursively:
            for result in op.results:
                for user in result.uses:
                    self._worklist.push(user.operation)

    def _get_rewriter_listener(self) -> PatternRewriterListener:
        """
        Get the listener that will be passed to the rewriter.
        It will take care of adding operations to the worklist, and calling the
        listener passed as configuration to the walker.
        """
        return PatternRewriterListener(
            operation_insertion_handler=[
                *self.listener.operation_insertion_handler,
                self._handle_operation_insertion,
            ],
            operation_removal_handler=[
                *self.listener.operation_removal_handler,
                self._handle_operation_removal,
            ],
            operation_modification_handler=[
                *self.listener.operation_modification_handler,
                self._handle_operation_modification,
            ],
            operation_replacement_handler=[
                *self.listener.operation_replacement_handler,
                self._handle_operation_replacement,
            ],
            block_creation_handler=self.listener.block_creation_handler,
        )

    def rewrite_module(self, module: ModuleOp) -> bool:
        """
        Rewrite operations nested in the given operation by repeatedly applying the
        pattern. Returns `True` if the IR was mutated.
        """
        return self.rewrite_op(module)

    def rewrite_op(self, op: Operation) -> bool:
        """
        Rewrite operations nested in the given operation by repeatedly applying the
        pattern. Returns `True` if the IR was mutated.
        """
        pattern_listener = self._get_rewriter_listener()

        self._populate_worklist(op)
        op_was_modified = self._process_worklist(pattern_listener)

        if not self.apply_recursively:
            return op_was_modified

        result = op_was_modified

        while op_was_modified:
            self._populate_worklist(op)
            op_was_modified = self._process_worklist(pattern_listener)

        return result

    def _populate_worklist(self, op: Operation) -> None:
        """Populate the worklist with all nested operations."""
        # We walk in reverse order since we use a stack for our worklist.
        for sub_op in op.walk(
            reverse=not self.walk_reverse, region_first=not self.walk_regions_first
        ):
            self._worklist.push(sub_op)

    def _process_worklist(self, listener: PatternRewriterListener) -> bool:
        """
        Process the worklist until it is empty.
        Returns true if any modification was done.
        """
        rewriter_has_done_action = False

        # Handle empty worklist
        op = self._worklist.pop()
        if op is None:
            return rewriter_has_done_action

        # Create a rewriter on the first operation
        rewriter = PatternRewriter(op)
        rewriter.extend_from_listener(listener)

        # do/while loop
        while True:
            # Reset the rewriter on `op`
            rewriter.has_done_action = False
            rewriter.current_operation = op

            # Apply the pattern on the operation
            try:
                self.pattern.match_and_rewrite(op, rewriter)
            except Exception as err:
                op.emit_error(
                    f"Error while applying pattern: {str(err)}",
                    exception_type=type(err),
                    underlying_error=err,
                )
            rewriter_has_done_action |= rewriter.has_done_action

            # If the worklist is empty, we are done
            op = self._worklist.pop()
            if op is None:
                return rewriter_has_done_action



from collections.abc import Mapping
from typing import cast

from setuptools import Command, find_packages, setup

import versioneer

setup(
    version=versioneer.get_version(),
    cmdclass=cast(Mapping[str, type[Command]], versioneer.get_cmdclass()),
    packages=find_packages(),
)



"""
This file contains a lexer and a parser for the MLIR declarative assembly format:
https://mlir.llvm.org/docs/DefiningDialects/Operations/#declarative-assembly-format
"""

from __future__ import annotations

import re
from collections.abc import Callable
from dataclasses import dataclass, field
from enum import Enum, auto
from itertools import pairwise
from typing import cast

from xdsl.dialects.builtin import Builtin
from xdsl.ir import Attribute, TypedAttribute
from xdsl.irdl import (
    AttrOrPropDef,
    AttrSizedOperandSegments,
    OpDef,
    OptionalDef,
    OptOperandDef,
    OptResultDef,
    ParamAttrConstraint,
    ParsePropInAttrDict,
    VariadicDef,
    VarOperandDef,
    VarResultDef,
)
from xdsl.irdl.declarative_assembly_format import (
    AnchorableDirective,
    AttrDictDirective,
    AttributeVariable,
    FormatDirective,
    FormatProgram,
    KeywordDirective,
    OperandOrResult,
    OperandTypeDirective,
    OperandVariable,
    OptionalAttributeVariable,
    OptionalGroupDirective,
    OptionallyParsableDirective,
    OptionalOperandTypeDirective,
    OptionalOperandVariable,
    OptionalResultTypeDirective,
    OptionalResultVariable,
    PunctuationDirective,
    ResultTypeDirective,
    ResultVariable,
    VariableDirective,
    VariadicLikeFormatDirective,
    VariadicLikeTypeDirective,
    VariadicLikeVariable,
    VariadicOperandTypeDirective,
    VariadicOperandVariable,
    VariadicResultTypeDirective,
    VariadicResultVariable,
    WhitespaceDirective,
)
from xdsl.parser import BaseParser, ParserState
from xdsl.utils.lexer import Input, Lexer, Token


@dataclass
class FormatLexer(Lexer):
    """
    A lexer for the declarative assembly format.
    The differences with the MLIR lexer are the following:
    * It can parse '`' or '$' as tokens. The token will have the `BARE_IDENT` kind.
    * Bare identifiers may also may contain `-`.
    """

    def lex(self) -> Token:
        """Lex a token from the input, and returns it."""
        # First, skip whitespaces
        self._consume_whitespace()

        start_pos = self.pos
        current_char = self._peek_chars()

        # Handle end of file
        if current_char is None:
            return self._form_token(Token.Kind.EOF, start_pos)

        # We parse '`', `\\` and '$' as a BARE_IDENT.
        # This is a hack to reuse the MLIR lexer.
        if current_char in ("`", "$", "\\", "^"):
            self._consume_chars()
            return self._form_token(Token.Kind.BARE_IDENT, start_pos)
        return super().lex()

    # Authorize `-` in bare identifier
    bare_identifier_suffix_regex = re.compile(r"[a-zA-Z0-9_$.\-]*")


class ParsingContext(Enum):
    """Indicates if the parser is nested in a particular directive."""

    TopLevel = auto()
    TypeDirective = auto()


@dataclass(init=False)
class FormatParser(BaseParser):
    """
    Parser for the declarative assembly format.
    The parser keeps track of the operands, operand types, results, regions, and
    attributes that are already parsed, and checks at the end that the provided format
    is correct, i.e., that it is unambiguous and refer to all elements exactly once.
    """

    op_def: OpDef
    """The operation definition we are parsing the format for."""
    seen_operands: list[bool]
    """The operand variables that are already parsed."""
    seen_operand_types: list[bool]
    """The operand types that are already parsed."""
    seen_result_types: list[bool]
    """The result types that are already parsed."""
    seen_attributes: set[str]
    """The attributes that are already parsed."""
    seen_properties: set[str]
    """The properties that are already parsed."""
    has_attr_dict: bool = field(default=False)
    """True if the attribute dictionary has already been parsed."""
    context: ParsingContext = field(default=ParsingContext.TopLevel)
    """Indicates if the parser is nested in a particular directive."""
    type_resolutions: dict[
        tuple[OperandOrResult, int],
        tuple[Callable[[Attribute], Attribute], OperandOrResult, int],
    ]
    """Map a variable to a way to infer its type"""

    def __init__(self, input: str, op_def: OpDef):
        super().__init__(ParserState(FormatLexer(Input(input, "<input>"))))
        self.op_def = op_def
        self.seen_operands = [False] * len(op_def.operands)
        self.seen_operand_types = [False] * len(op_def.operands)
        self.seen_result_types = [False] * len(op_def.results)
        self.seen_attributes = set[str]()
        self.seen_properties = set[str]()
        self.type_resolutions = {}

    def parse_format(self) -> FormatProgram:
        """
        Parse a declarative format, with the following syntax:
          format ::= directive*
        Once the format is parsed, check that it is correct, i.e., that it is
        unambiguous and refer to all elements exactly once.
        """
        elements: list[FormatDirective] = []
        while self._current_token.kind != Token.Kind.EOF:
            elements.append(self.parse_directive())

        self.add_reserved_attrs_to_directive(elements)
        seen_variables = self.resolve_types()
        self.verify_directives(elements)
        self.verify_attr_dict()
        self.verify_properties()
        self.verify_operands(seen_variables)
        self.verify_results(seen_variables)
        return FormatProgram(elements)

    def verify_directives(self, elements: list[FormatDirective]):
        """
        Check correctness of the declarative format; e.g, chaining variadiclike operand
        directives leads to ambiguous parsing, and should raise an error here.
        """
        for a, b in pairwise(elements):
            match a, b:
                case VariadicLikeFormatDirective(), PunctuationDirective(","):
                    self.raise_error(
                        "A variadic directive cannot be followed by a comma literal."
                    )
                case VariadicLikeTypeDirective(), VariadicLikeTypeDirective():
                    self.raise_error(
                        "A variadic type directive cannot be followed by another variadic type directive."
                    )
                case VariadicLikeVariable(), VariadicLikeVariable() if not (
                    isinstance(a, VariadicLikeTypeDirective)
                    or isinstance(b, VariadicLikeTypeDirective)
                ):
                    self.raise_error(
                        "A variadic operand variable cannot be followed by another variadic operand variable."
                    )
                case _:
                    pass

    def add_reserved_attrs_to_directive(self, elements: list[FormatDirective]):
        """
        Add reserved attributes to the attr-dict directive.
        These are the attributes that are printed/parsed in other places in the format,
        and thus should not be printed in the attr-dict directive.
        """
        for idx, element in enumerate(elements):
            if isinstance(element, AttrDictDirective):
                elements[idx] = AttrDictDirective(
                    with_keyword=element.with_keyword,
                    reserved_attr_names=self.seen_attributes,
                    print_properties=element.print_properties,
                )
                return

    def resolve_types(self) -> set[str]:
        """
        Find out which constraint variables can be inferred from the parsed attributes.
        """
        seen_variables = set[str]()
        for i, (_, operand_def) in enumerate(self.op_def.operands):
            if self.seen_operand_types[i]:
                seen_variables |= operand_def.constr.get_resolved_variables()
        for i, (_, result_def) in enumerate(self.op_def.results):
            if self.seen_result_types[i]:
                seen_variables |= result_def.constr.get_resolved_variables()
        return seen_variables

    def verify_operands(self, seen_variables: set[str]):
        """
        Check that all operands and operand types are refered at least once, or inferred
        from another construct.
        """
        for (
            seen_operand,
            seen_operand_type,
            (operand_name, operand_def),
        ) in zip(
            self.seen_operands,
            self.seen_operand_types,
            self.op_def.operands,
            strict=True,
        ):
            if not seen_operand:
                self.raise_error(
                    f"operand '{operand_name}' "
                    f"not found, consider adding a '${operand_name}' "
                    "directive to the custom assembly format"
                )
            if not seen_operand_type:
                if not operand_def.constr.can_infer(seen_variables):
                    self.raise_error(
                        f"type of operand '{operand_name}' cannot be inferred, "
                        f"consider adding a 'type(${operand_name})' directive to the "
                        "custom assembly format"
                    )

    def verify_results(self, seen_variables: set[str]):
        """Check that all result types are refered at least once, or inferred
        from another construct."""

        for result_type, (result_name, result_def) in zip(
            self.seen_result_types, self.op_def.results, strict=True
        ):
            if not result_type:
                if not result_def.constr.can_infer(seen_variables):
                    self.raise_error(
                        f"type of result '{result_name}' cannot be inferred, "
                        f"consider adding a 'type(${result_name})' directive to the "
                        "custom assembly format"
                    )

    def verify_attr_dict(self):
        """
        Check that the attribute dictionary is present.
        """
        if not self.has_attr_dict:
            self.raise_error("'attr-dict' directive not found")

    def verify_properties(self):
        """
        Check that all properties are present, unless `ParsePropInAttrDict` option is
        used.
        """
        # This is used for compatibility with MLIR
        if any(
            isinstance(option, ParsePropInAttrDict) for option in self.op_def.options
        ):
            if self.seen_properties:
                self.raise_error(
                    "properties cannot be specified in the declarative format "
                    "when 'ParsePropInAttrDict' IRDL option is used. They are instead "
                    "parsed from the attribute dictionary."
                )
            return
        missing_properties = set(self.op_def.properties.keys()) - self.seen_properties
        if missing_properties:
            self.raise_error(
                f"{', '.join(missing_properties)} properties are missing from "
                "the declarative format. If this is intentional, consider using "
                "'ParsePropInAttrDict' IRDL option."
            )

    def parse_optional_variable(
        self,
    ) -> VariableDirective | AttributeVariable | None:
        """
        Parse a variable, if present, with the following format:
          variable ::= `$` bare-ident
        The variable should refer to an operand, attribute, region, result,
        or successor.
        """
        if self._current_token.text[0] != "$":
            return None
        self._consume_token()
        variable_name = self.parse_identifier(" after '$'")

        # Check if the variable is an operand
        for idx, (operand_name, operand_def) in enumerate(self.op_def.operands):
            if variable_name != operand_name:
                continue
            if self.context == ParsingContext.TopLevel:
                if self.seen_operands[idx]:
                    self.raise_error(f"operand '{variable_name}' is already bound")
                self.seen_operands[idx] = True
                if isinstance(operand_def, VariadicDef | OptionalDef):
                    self.seen_attributes.add(AttrSizedOperandSegments.attribute_name)
            match operand_def:
                case OptOperandDef():
                    return OptionalOperandVariable(variable_name, idx)
                case VarOperandDef():
                    return VariadicOperandVariable(variable_name, idx)
                case _:
                    return OperandVariable(variable_name, idx)

        # Check if the variable is a result
        for idx, (result_name, result_def) in enumerate(self.op_def.results):
            if variable_name != result_name:
                continue
            if self.context == ParsingContext.TopLevel:
                self.raise_error(
                    "result variable cannot be in a toplevel directive. "
                    f"Consider using 'type({variable_name})' instead."
                )
            match result_def:
                case OptResultDef():
                    return OptionalResultVariable(variable_name, idx)
                case VarResultDef():
                    return VariadicResultVariable(variable_name, idx)
                case _:
                    return ResultVariable(variable_name, idx)
            if isinstance(result_def, VariadicDef):
                return VariadicResultVariable(variable_name, idx)
            else:
                return ResultVariable(variable_name, idx)

        attr_or_prop_by_name = {
            attr_name: attr_or_prop
            for attr_name, attr_or_prop in self.op_def.accessor_names.values()
        }

        # Check if the variable is an attribute
        if variable_name in attr_or_prop_by_name:
            attr_name = variable_name
            attr_or_prop = attr_or_prop_by_name[attr_name]
            if self.context == ParsingContext.TopLevel:
                if attr_or_prop == "property":
                    if attr_name in self.seen_properties:
                        self.raise_error(f"property '{variable_name}' is already bound")
                    self.seen_properties.add(attr_name)
                else:
                    if attr_name in self.seen_attributes:
                        self.raise_error(
                            f"attribute '{variable_name}' is already bound"
                        )
                    self.seen_attributes.add(attr_name)

            attr_def = (
                self.op_def.properties.get(attr_name)
                if attr_or_prop == "property"
                else self.op_def.attributes.get(attr_name)
            )
            if isinstance(attr_def, AttrOrPropDef):
                unique_base = attr_def.constr.get_unique_base()
                # Always qualify builtin attributes
                # This is technically an approximation, but appears to be good enough
                # for xDSL right now.
                unique_type = None
                if unique_base is not None and issubclass(unique_base, TypedAttribute):
                    constr = attr_def.constr
                    # TODO: generalize.
                    # https://github.com/xdslproject/xdsl/issues/2499
                    if isinstance(constr, ParamAttrConstraint):
                        type_constraint = constr.param_constrs[
                            unique_base.get_type_index()
                        ]
                        if type_constraint.can_infer(set()):
                            unique_type = type_constraint.infer({})
                if (
                    unique_base is not None
                    and unique_base in Builtin.attributes
                    and unique_type is None
                ):
                    unique_base = None

                # Chill pyright with TypedAttribute without parameter
                unique_base = cast(type[Attribute] | None, unique_base)

                variable_type = (
                    OptionalAttributeVariable
                    if isinstance(attr_def, OptionalDef)
                    else AttributeVariable
                )
                is_property = attr_or_prop == "property"
                return variable_type(
                    variable_name, is_property, unique_base, unique_type
                )

        self.raise_error(
            "expected variable to refer to an operand, "
            "attribute, region, result, or successor"
        )

    def parse_type_directive(self) -> FormatDirective:
        """
        Parse a type directive with the following format:
          type-directive ::= `type` `(` variable `)`
        `type` is expected to have already been parsed
        """
        self.parse_punctuation("(")

        # Update the current context, since we are now in a type directive
        previous_context = self.context
        self.context = ParsingContext.TypeDirective

        variable = self.parse_optional_variable()
        match variable:
            case None:
                self.raise_error("'type' directive expects a variable argument")
            case OptionalOperandVariable(name, index):
                if self.seen_operand_types[index]:
                    self.raise_error(f"types of '{name}' is already bound")
                self.seen_operand_types[index] = True
                res = OptionalOperandTypeDirective(name, index)
            case VariadicOperandVariable(name, index):
                if self.seen_operand_types[index]:
                    self.raise_error(f"types of '{name}' is already bound")
                self.seen_operand_types[index] = True
                res = VariadicOperandTypeDirective(name, index)
            case OperandVariable(name, index):
                if self.seen_operand_types[index]:
                    self.raise_error(f"type of '{name}' is already bound")
                self.seen_operand_types[index] = True
                res = OperandTypeDirective(name, index)
            case OptionalResultVariable(name, index):
                if self.seen_result_types[index]:
                    self.raise_error(f"types of '{name}' is already bound")
                self.seen_result_types[index] = True
                res = OptionalResultTypeDirective(name, index)
            case VariadicResultVariable(name, index):
                if self.seen_result_types[index]:
                    self.raise_error(f"types of '{name}' is already bound")
                self.seen_result_types[index] = True
                res = VariadicResultTypeDirective(name, index)
            case ResultVariable(name, index):
                if self.seen_result_types[index]:
                    self.raise_error(f"type of '{name}' is already bound")
                self.seen_result_types[index] = True
                res = ResultTypeDirective(name, index)
            case AttributeVariable():
                self.raise_error("can only take the type of an operand or result")
            case _:
                raise ValueError(f"Unexpected variable type {type(variable)}")

        self.parse_punctuation(")")
        self.context = previous_context
        return res

    def parse_optional_group(self) -> FormatDirective:
        """
        Parse an optional group, with the following format:
          group ::= `(` then-elements `)` `?`
        """
        then_elements = tuple[FormatDirective, ...]()
        anchor: FormatDirective | None = None

        while not self.parse_optional_punctuation(")"):
            then_elements += (self.parse_directive(),)
            if self.parse_optional_keyword("^"):
                if anchor is not None:
                    self.raise_error("An optional group can only have one anchor.")
                anchor = then_elements[-1]
        self.parse_punctuation("?")

        if not then_elements:
            self.raise_error("An optional group cannot be empty")
        if anchor is None:
            self.raise_error("Every optional group must have an anchor.")
        # TODO: allow attribute and region variables when implemented.
        if not isinstance(then_elements[0], OptionallyParsableDirective):
            self.raise_error(
                "First element of an optional group must be optionally parsable."
            )
        if not isinstance(anchor, AnchorableDirective):
            self.raise_error(
                "An optional group's anchor must be an achorable directive."
            )

        return OptionalGroupDirective(anchor, then_elements[0], then_elements[1:])

    def parse_keyword_or_punctuation(self) -> FormatDirective:
        """
        Parse a keyword or a punctuation directive, with the following format:
          keyword-or-punctuation-directive ::= `\\`` (bare-ident | punctuation) `\\``
        """
        self.parse_characters("`")
        start_token = self._current_token

        # New line case
        if self.parse_optional_keyword("\\"):
            self.parse_keyword("n")
            self.parse_characters("`")
            return WhitespaceDirective("\n")

        # Space case
        if self.parse_optional_characters("`"):
            end_token = self._current_token
            whitespace = self.lexer.input.content[
                start_token.span.end : end_token.span.start
            ]
            if whitespace != " ":
                self.raise_error(
                    "unexpected whitespace in directive, only ` ` whitespace is allowed"
                )
            return WhitespaceDirective(" ")

        # Punctuation case
        if self._current_token.kind.is_punctuation():
            punctuation = self._consume_token().text
            self.parse_characters("`")
            assert Token.Kind.is_spelling_of_punctuation(punctuation)
            return PunctuationDirective(punctuation)

        # Identifier case
        ident = self.parse_optional_identifier()
        if ident is None or ident == "`":
            self.raise_error("punctuation or identifier expected")

        self.parse_characters("`")
        return KeywordDirective(ident)

    def parse_directive(self) -> FormatDirective:
        """
        Parse a format directive, with the following format:
          directive ::= `attr-dict`
                        | `attr-dict-with-keyword`
                        | type-directive
                        | keyword-or-punctuation-directive
                        | variable
        """
        if self.parse_optional_keyword("attr-dict"):
            return self.create_attr_dict_directive(False)
        if self.parse_optional_keyword("attr-dict-with-keyword"):
            return self.create_attr_dict_directive(True)
        if self.parse_optional_keyword("type"):
            return self.parse_type_directive()
        if self._current_token.text == "`":
            return self.parse_keyword_or_punctuation()
        if self.parse_optional_punctuation("("):
            return self.parse_optional_group()
        if variable := self.parse_optional_variable():
            return variable
        self.raise_error(f"unexpected token '{self._current_token.text}'")

    def create_attr_dict_directive(self, with_keyword: bool) -> AttrDictDirective:
        """Create an attribute dictionary directive, and update the parsing state."""
        if self.has_attr_dict:
            self.raise_error(
                "'attr-dict' directive can only occur once "
                "in the assembly format description"
            )
        self.has_attr_dict = True
        print_properties = any(
            isinstance(option, ParsePropInAttrDict) for option in self.op_def.options
        )
        # reserved_attr_names is populated once the format is parsed, as some attributes
        # might appear after the attr-dict directive
        return AttrDictDirective(
            with_keyword=with_keyword,
            reserved_attr_names=set(),
            print_properties=print_properties,
        )



from collections.abc import Hashable
from dataclasses import dataclass

from xdsl.dialects.builtin import ModuleOp
from xdsl.utils.hasher import Hasher


@dataclass(frozen=True)
class HashableModule(Hashable):
    """
    A class wrapping a ModuleOp, that forwards to structural equality checking for `==`.
    The module in this class should not be mutated.
    """

    module: ModuleOp

    def __eq__(self, other: object) -> bool:
        return isinstance(
            other, HashableModule
        ) and self.module.is_structurally_equivalent(other.module)

    def __hash__(self) -> int:
        """
        The hash of the module is a hash of the ordered combination of operation names.
        As most transformations on IR modify at least one operation, this should be enough
        to minimise collisions.
        """
        hasher = Hasher()
        for op in self.module.walk():
            hasher.combine(op.name)
        return hasher.hash



from onnx import GraphProto, NodeProto, ValueInfoProto

from xdsl.dialects import func, onnx
from xdsl.dialects.builtin import ModuleOp
from xdsl.frontend.onnx.type import get_type
from xdsl.ir import Attribute, SSAValue
from xdsl.irdl import IRDLOperation

OP_BY_OP_TYPE: dict[str, type[IRDLOperation]] = {
    "Add": onnx.Add,
    "Sub": onnx.Sub,
    "MatMul": onnx.MatMul,
    "Transpose": onnx.Transpose,
}
"""Associate the name of the operations with the respective operation in ONNX dialect."""


class OnnxXdslMapping:
    """The representation of the onnx context."""

    type_by_name: dict[str, Attribute]
    value_by_name: dict[str, SSAValue]

    def __init__(self):
        self.type_by_name = {}
        self.value_by_name = {}


def visit_node(node: NodeProto, ctx: OnnxXdslMapping) -> IRDLOperation:
    """Update the onnx context with the current node of the onnx graph."""
    if node.op_type not in OP_BY_OP_TYPE:
        raise ValueError(f"Unknown ONNX op name {node.op_type}")

    op_class = OP_BY_OP_TYPE[node.op_type]

    operands = tuple(ctx.value_by_name[name] for name in node.input)
    result_types = tuple(ctx.type_by_name[name] for name in node.output)

    op = op_class.build(operands=operands, result_types=result_types)
    results = op.results

    for output_name, result in zip(node.output, results, strict=True):
        ctx.value_by_name[output_name] = result

    return op


def visit_value_info(i: ValueInfoProto, ctx: OnnxXdslMapping) -> Attribute:
    """Given the onnx ValueInforProto, it returns the corresponding Attribute stored in the context."""
    name = i.name

    if name in ctx.type_by_name:
        return ctx.type_by_name[name]

    t = get_type(i.type)
    ctx.type_by_name[name] = t
    return t


def build_module(graph: GraphProto) -> ModuleOp:
    """Create the ModuleOp based on the onnx graph provided."""

    ctx = OnnxXdslMapping()
    fn = visit_graph(graph, ctx)
    module = ModuleOp([fn])

    return module


def visit_graph(g: GraphProto, ctx: OnnxXdslMapping) -> IRDLOperation:
    """
    Visit the onnx graph to update the onnx context.
    """

    name = g.name

    input_types = tuple(visit_value_info(input, ctx) for input in g.input)
    output_types = tuple(visit_value_info(output, ctx) for output in g.output)

    fn = func.FuncOp(name, (input_types, output_types))

    for input, arg in zip(g.input, fn.body.block.args, strict=True):
        ctx.value_by_name[input.name] = arg

    for node in g.node:
        results = visit_node(node, ctx)
        fn.body.block.add_op(results)

    returned_values = tuple(ctx.value_by_name[output.name] for output in g.output)
    retfn = func.Return(*returned_values)
    fn.body.block.add_op(retfn)

    return fn



from __future__ import annotations

from xdsl.dialects.builtin import Float32Type, Float64Type
from xdsl.ir import (
    Dialect,
    Operation,
    OpResult,
    ParametrizedAttribute,
    SSAValue,
    TypeAttribute,
)
from xdsl.irdl import (
    AnyOf,
    IRDLOperation,
    Operand,
    ParamAttrConstraint,
    ParameterDef,
    irdl_attr_definition,
    irdl_op_definition,
    operand_def,
    result_def,
)
from xdsl.utils.exceptions import VerifyException


@irdl_attr_definition
class ComplexType(ParametrizedAttribute, TypeAttribute):
    name = "cmath.complex"
    data: ParameterDef[Float64Type | Float32Type]


@irdl_op_definition
class Norm(IRDLOperation):
    name = "cmath.norm"

    op: Operand = operand_def(
        ParamAttrConstraint(ComplexType, [AnyOf([Float32Type, Float64Type])])
    )
    res: OpResult = result_def(AnyOf([Float32Type, Float64Type]))

    # TODO replace with trait
    def verify_(self) -> None:
        if not isinstance(self.op.type, ComplexType):
            raise VerifyException("Expected complex type")
        if self.op.type.data != self.res.type:
            raise VerifyException("expect all input and output types to be equal")


@irdl_op_definition
class Mul(IRDLOperation):
    name = "cmath.mul"

    lhs: Operand = operand_def(
        ParamAttrConstraint(ComplexType, [AnyOf([Float32Type, Float64Type])])
    )
    rhs: Operand = operand_def(
        ParamAttrConstraint(ComplexType, [AnyOf([Float32Type, Float64Type])])
    )
    result: OpResult = result_def(
        ParamAttrConstraint(ComplexType, [AnyOf([Float32Type, Float64Type])])
    )

    # TODO replace with trait
    def verify_(self) -> None:
        if self.lhs.type != self.rhs.type and self.rhs.type != self.result.type:
            raise VerifyException("expect all input and output types to be equal")

    @staticmethod
    def get(operand1: Operation | SSAValue, operand2: Operation | SSAValue) -> Mul:
        operand1 = SSAValue.get(operand1)
        return Mul.build(operands=[operand1, operand2], result_types=[operand1.type])


CMath = Dialect("cmath", [Norm, Mul], [ComplexType])
